#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test029_volume_time_rotation_helpers as test029
import opengate as gate
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test029"
    )

    # create the main simulation object
    sim = gate.Simulation()

    # create sim without AA
    test029.create_simulation(sim, True, paths, "_trot2")

    # for later reference, get the actors that were created by the helper function above
    proj_actor = sim.actor_manager.get_actor("Projection")
    stats = sim.actor_manager.get_actor("Stats")

    # initialize & start
    sim.run()

    """
    WARNING when "angle_acceptance_volume" is enabled, it is a bit faster (+50%) but the result is not
    exactly the same as without. This is because, even if the initial particle is not in the direction of
    the spect system, it can scatter and still reach the detector.

    We don't have the collimator here (to accelerate the simulation), this is why the difference is not negligible.

    The (fake) reference is computed by scaling the ref by 90% (test029_volume_time_rotation_1.py)
    """

    gate.exception.warning("Compare stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "stats029.txt")
    print(
        f"Number of steps was {stats.counts.steps}, forced to the same value (because of angle acceptance). "
    )
    stats.counts.steps = stats_ref.counts.steps  # force these to be identical
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.01)
    print(is_ok)

    gate.exception.warning("Compare images")
    # read image and force change the offset to be similar to old Gate
    is_ok = (
        utility.assert_images(
            paths.output_ref / "proj029_scaled.mhd",
            proj_actor.get_output_path("counts"),
            stats,
            tolerance=60,
            ignore_value_data2=0,
            axis="x",
            sum_tolerance=6,
            apply_ignore_mask_to_sum_check=False,  # reproduce legacy behavior
        )
        and is_ok
    )
    print(is_ok)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, None, "test054")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu_type = "vrml"
    sim.visu = False
    sim.check_volumes_overlap = True
    sim.number_of_threads = 1
    sim.random_seed = 654923
    sim.output_dir = paths.output

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # image
    patient = sim.add_volume("Image", "patient")
    patient.mother = fake.name
    patient.image = paths.data / "patient-40mm.mhd"
    patient.material = "G4_AIR"  # material used by default
    patient.voxel_materials = [
        [-2000, -900, "G4_AIR"],
        [-900, -100, "Lung"],
        [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
        [0, 300, "G4_TISSUE_SOFT_ICRP"],
        [300, 800, "G4_B-100_BONE"],
        [800, 6000, "G4_BONE_COMPACT_ICRU"],
    ]

    # create two other parallel worlds
    sim.add_parallel_world("world2")
    sim.add_parallel_world("world3")

    # detector in w2 (on top of world)
    det = sim.add_volume("Box", "detector")
    det.mother = "world2"
    det.material = "G4_GLASS_LEAD"
    det.size = [400 * mm, 400 * mm, 2 * mm]
    det.translation = [0, 0, 200 * mm]
    det.rotation = Rotation.from_euler("x", 40, degrees=True).as_matrix()
    det.color = [1, 0, 0, 1]

    # detector in w3 (on top of w2)
    det2 = sim.add_volume("Box", "detector2")
    det2.mother = "world3"
    det2.material = "G4_GLASS_LEAD"  # set 'None' if this volume should be transparent
    det2.size = [400 * mm, 400 * mm, 2 * mm]
    det2.translation = [0, 0, 200 * mm]
    det2.rotation = Rotation.from_euler("x", -40, degrees=True).as_matrix()
    det2.color = [1, 0, 0, 1]

    # source
    source = sim.add_source("GenericSource", "source")
    source.energy.mono = 0.1 * MeV
    source.particle = "gamma"
    source.position.type = "box"
    source.position.size = [1 * mm, 200 * mm, 1 * mm]
    source.position.translation = [0, 0, 0 * cm]
    source.activity = 10000 * Bq / sim.number_of_threads
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]

    # add phsp actor detector 1 (overlap!)
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.output_filename = "test054.root"
    phsp.attached_to = det.name
    phsp.attributes = [
        "KineticEnergy",
        "PrePosition",
    ]
    phsp.steps_to_store = "exiting first"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # print
    print("Geometry trees: ")
    print(sim.volume_manager.dump_volume_tree())

    # start simulation
    sim.run(True)

    # print results at the end
    print(stats)

    keys = ["KineticEnergy", "PrePosition_X", "PrePosition_Y", "PrePosition_Z"]
    tols = [0.01, 2.6, 1.8, 1.7]
    ref = paths.output_ref / "test054_ref.root"
    f = paths.output / "test054.png"
    is_ok = utility.compare_root3(
        ref,
        phsp.get_output_path(),
        "phsp",
        "phsp",
        keys,
        keys,
        tols,
        None,
        None,
        f,
        hits_tol=6.1,
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.image import get_translation_to_isocenter, resample_itk_image_like
from opengate.tests import utility
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.voxelize import (
    voxelize_geometry,
    write_voxelized_geometry,
    voxelized_source,
)
import itk
import json

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test015")

    # create the simulation
    sim = gate.Simulation()
    # sim.visu = True
    # sim.visu_type = "qt"
    sim.random_seed = 87456321
    sim.output_dir = paths.output
    sim.progress_bar = True

    # units
    MeV = gate.g4_units.MeV
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 10 * mm)

    # world size
    sim.world.size = [2 * m, 2 * m, 2 * m]

    # add an iec phantom
    iec = gate_iec.add_iec_phantom(sim)

    # add sources
    sources = []
    a = 100 * BqmL
    if sim.visu:
        a = 1 * BqmL
    s = gate_iec.add_central_cylinder_source(sim, iec.name, "s_cyl", a / 2)
    sources.append(s)
    s = gate_iec.add_spheres_sources(sim, iec.name, "s_sphere", "all", [a] * 6)
    for b in s:
        sources.append(b)
    s = gate_iec.add_background_source(sim, iec.name, "s_bg", a / 20)
    sources.append(s)

    total_activity = 0
    for s in sources:
        s.particle = "alpha"
        s.energy.type = "mono"
        s.energy.mono = 100 * MeV
        total_activity += s.activity
    print(f"Total activity: {total_activity/Bq} Bq")

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True
    stats.output_filename = "test015_iec_phantom_5_stats.txt"

    # add dose actor
    dose1 = sim.add_actor("DoseActor", "dose1")
    dose1.edep.output_filename = "test015_iec_5_edep.mhd"
    dose1.attached_to = iec
    dose1.size = [150, 150, 150]
    dose1.spacing = [3 * mm, 3 * mm, 3 * mm]
    dose1.output_coordinate_system = "local"

    # voxelized iec
    spacing = (3.5 * mm, 2 * mm, 4.5 * mm)
    if sim.visu:
        spacing = (8 * mm, 8 * mm, 8 * mm)
    volume_labels, image = voxelize_geometry(sim, extent=iec, spacing=spacing, margin=1)
    filenames = write_voxelized_geometry(
        sim, volume_labels, image, paths.output / "test015_iec_5_geom.mhd"
    )
    print(filenames)

    # voxelize source
    a = 1
    activities = {
        "iec_sphere_10mm": a,
        "iec_sphere_13mm": a,
        "iec_sphere_17mm": a,
        "iec_sphere_22mm": a,
        "iec_sphere_28mm": a,
        "iec_sphere_37mm": a,
        "iec_interior": a / 20,
        "iec_center_cylinder_hole": a / 2,
    }

    img = itk.imread(filenames["image"])
    volume_labels = json.loads(open(filenames["volumes"]).read())
    img_source = voxelized_source(img, volume_labels, activities)
    itk.imwrite(img_source, paths.output / "test015_iec_5_geom_source.mhd")

    # add voxelized IEC
    # Warning: the position must be adapted for IEC (default position is center of the image)
    vox = sim.add_volume("ImageVolume", "iec_vox")
    vox.image = filenames["image"]
    vox.read_label_to_material(filenames["labels"])
    vox.translation = get_translation_to_isocenter(vox.image)
    # move on the left to avoid overlap
    vox.translation[0] -= 35 * cm

    # add source by reading vox
    source = sim.add_source("VoxelSource", "source")
    source.attached_to = vox
    source.image = paths.output / "test015_iec_5_geom_source.mhd"
    source.particle = "alpha"
    source.energy.type = "mono"
    source.energy.mono = 100 * MeV
    source.activity = total_activity

    # add another dose actor
    dose2 = sim.add_actor("DoseActor", "dose_vox")
    dose2.edep.output_filename = "test015_iec_5_edep_vox.mhd"
    dose2.attached_to = vox
    dose2.size = [150, 150, 150]
    dose2.spacing = [3 * mm, 3 * mm, 3 * mm]
    dose2.output_coordinate_system = "attached_to_image"

    # start
    sim.run()
    print(stats)

    # resample images in the same coordinate system
    img1 = itk.imread(dose1.edep.get_output_path())
    img2 = itk.imread(dose2.edep.get_output_path())
    img2_res = resample_itk_image_like(img2, img1, 0, linear=True)
    fn = str(dose2.edep.get_output_path()).replace(".mhd", "_resampled.mhd")
    itk.imwrite(img2_res, fn)

    # compare images
    is_ok = utility.assert_images(
        dose1.get_output_path("edep"),
        fn,
        stats=None,
        axis="y",
        tolerance=150,
        sum_tolerance=1.1,
        sad_profile_tolerance=4.0,
        ignore_value_data1=0,
        ignore_value_data2=0,
        # apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test040")

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3
    sec = gate.g4_units.second
    deg = gate.g4_units.deg
    kBq = 1000 * Bq
    MBq = 1000 * kBq

    # main parameters
    sim.output_dir = paths.output
    sim.check_volumes_overlap = True
    sim.number_of_threads = 1
    sim.random_seed = 1386
    sim.output_dir = paths.output
    ac = 5e3 * BqmL / sim.number_of_threads
    sim.visu = False
    sim.visu_type = "vrml"
    colli_flag = not sim.visu
    if sim.visu:
        ac = 1 * BqmL  # per mL
        sim.number_of_threads = 1

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # iec phantom
    iec_phantom = gate_iec.add_iec_phantom(sim)

    # c, v = gate_iec.compute_sphere_centers_and_volumes(sim, iec_phantom.name)
    # print(c)
    # print(v)

    """# add an artificial tungsten bar
    vint = sim.volume_manager.volumes["iec_interior"]
    print(vint)
    t = sim.add_volume("Box", "tung")
    t.mother = vint.name
    t.size = [3 * cm, 7 * cm, 10 * cm]
    t.translation = [-10 * cm, 5 * cm, 2 * cm]
    t.material = "G4_CADMIUM_TUNGSTATE"
    t.color = [0, 0, 1, 1]"""

    # test phase space
    phsp_sphere_surface = sim.add_volume("Sphere", "phase_space_sphere")
    phsp_sphere_surface.rmin = 215 * mm
    phsp_sphere_surface.rmax = 216 * mm
    phsp_sphere_surface.color = [1, 1, 1, 1]
    phsp_sphere_surface.material = "G4_AIR"

    # physic list
    sim.physics_manager.set_production_cut("world", "all", 1 * mm)

    # source sphere
    gate_iec.add_spheres_sources(
        sim,
        "iec",
        "source1",
        [10, 13, 17, 22, 28, 37],
        [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac],
        verbose=True,
    )

    # with acceptance angle (?) # FIXME
    sources = sim.source_manager.sources
    for source in sources.values():
        source.particle = "e+"
        source.energy.type = "Ga68"
        source.direction.type = "iso"

    # background source 1:10 ratio with sphere
    # bg = gate_iec.add_background_source(sim, 'iec', 'source_bg', ac / 10, verbose=True)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.output_filename = "test040_ref_stats.txt"

    # store phsp of exiting particles (gamma only)
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.attached_to = phsp_sphere_surface.name
    phsp.attributes = [
        "KineticEnergy",
        "PrePosition",
        "PreDirection",
        "GlobalTime",
        "TimeFromBeginOfEvent",
        "EventID",
        "EventPosition",
        "EventDirection",
        "EventKineticEnergy",
    ]
    phsp.output_filename = "test040_ref_phsp.root"
    phsp.store_absorbed_event = True
    # phsp.steps_to_store = "first"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    phsp.filters.append(f)
    f = sim.add_filter("KineticEnergyFilter", "f2")
    f.energy_min = 100 * keV
    phsp.filters.append(f)

    # go
    sim.run()

    # ----------------------------------------------------------------------------------------------------------

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "test040_ref_stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.01)

    # 426760*2*0.8883814158496728 = 758251.3
    ref = 9523
    ae = phsp.number_of_absorbed_events
    print(ae)
    err = abs(ae - ref) / ref
    tol = 0.055
    is_ok = err < tol and is_ok
    utility.print_test(
        is_ok, f"Number of absorbed events: {ae} vs {ref} = {err * 100:.2f}%"
    )

    # No other tests here for the moment, will be used by test040_gan
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from opengate.devtools import check_if_class_has_attribute, apply_class_check_to_package
from opengate.base import GateObject
from opengate.exception import GateImplementationError
import opengate.tests.utility as utility

if __name__ == "__main__":
    is_ok = True
    exceptions = []

    # check if all actors implement an __initcpp__ method as they should
    attribute_name = "__initcpp__"
    warnings = apply_class_check_to_package(
        check_if_class_has_attribute,
        package_name="opengate",
        sub_package_name="actors",
        inherits_from="opengate.actors.base.ActorBase",
        func_kwargs={"attribute_name": attribute_name, "attribute_type": "method"},
    )
    if len(warnings):
        is_ok = False
        s = "\n".join([f"{i}) {w}" for i, w in enumerate(warnings)])
        exceptions.append(
            GateImplementationError(
                f"{len(warnings)} GateObjects do not implement a '{attribute_name}' method: \n{s}"
            )
        )

    # test the check function by performing a check that should fail!
    warnings_test = apply_class_check_to_package(
        check_if_class_has_attribute,
        package_name="opengate",
        sub_package_name="actors",
        inherits_from="opengate.actors.base.ActorBase",
        func_kwargs={"attribute_name": attribute_name, "attribute_type": "property"},
    )
    # so there should be warnings!
    if len(warnings_test) == 0:
        is_ok = False
        exceptions.append(
            GateImplementationError(
                "A class check should have issued a warning, but it has not. "
                "The class check looked for actor classes "
                "which do not implement an '__initcpp__ PROPERTY', and indeed "
                "no actor class should really do that. "
            )
        )

    utility.test_ok(is_ok, exceptions=exceptions)



# N.B: distances in mm, degrees in rad
# FIXME: make sure distances are correctly converted via g4units


class BeamlineModel:
    def __init__(self):
        self.name = None
        self.radiation_types = []
        self.rm = None  # range modulator
        self.rashi = None
        # Nozzle entrance to Isocenter distance
        self.distance_nozzle_iso = 0  # mm
        # SMX (X bending magnet) to Isocenter distance
        self.distance_stearmag_to_isocenter_x = float(
            "inf"
        )  # default infinity for parallel beams
        # SMY (Y bending magnet) to Isocenter distance
        self.distance_stearmag_to_isocenter_y = float("inf")
        # polinomial coefficients
        self.energy_mean_coeffs = [0]
        self.energy_spread_coeffs = [0]
        self.sigma_x_coeffs = [0]
        self.theta_x_coeffs = [0]
        self.epsilon_x_coeffs = [0]
        self.sigma_y_coeffs = [0]
        self.theta_y_coeffs = [0]
        self.epsilon_y_coeffs = [0]
        # convergence
        self.conv_x = 0
        self.conv_y = 0

    def _polynomial_map(self, base, coeff):
        # coeff are given with decreasing degree (coeff[0]->max degree)
        polyDegree = len(coeff)
        exp = list(range(polyDegree))
        exp.reverse()

        return sum([c * (base ** (i)) for c, i in zip(coeff, exp)])

    def get_energy(self, nominal_energy):
        return self._polynomial_map(nominal_energy, self.energy_mean_coeffs)

    def get_sigma_energy(self, nominal_energy):
        return self._polynomial_map(nominal_energy, self.energy_spread_coeffs)

    def get_sigma_x(self, energy):
        return self._polynomial_map(energy, self.sigma_x_coeffs)

    def get_theta_x(self, energy):
        return self._polynomial_map(energy, self.theta_x_coeffs)

    def get_epsilon_x(self, energy):
        return self._polynomial_map(energy, self.epsilon_x_coeffs)

    def get_sigma_y(self, energy):
        return self._polynomial_map(energy, self.sigma_y_coeffs)

    def get_theta_y(self, energy):
        return self._polynomial_map(energy, self.theta_y_coeffs)

    def get_epsilon_y(self, energy):
        return self._polynomial_map(energy, self.epsilon_y_coeffs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt

if __name__ == "__main__":
    # ------ INITIALIZE SIMULATION ENVIRONMENT ----------
    paths = utility.get_default_test_paths(__file__, "gate_test044_pbs", "test059")
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    s = gate.g4_units.s

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    ## ---------- DEFINE BEAMLINE MODEL -------------##
    beamline = BeamlineModel()
    beamline.name = None
    beamline.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    beamline.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    beamline.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    beamline.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    beamline.energy_mean_coeffs = [11.91893485094217, -9.539517997860457]
    beamline.energy_spread_coeffs = [0.0004790681841295621, 5.253257865904452]
    beamline.sigma_x_coeffs = [2.3335753978880014]
    beamline.theta_x_coeffs = [0.0002944903217664001]
    beamline.epsilon_x_coeffs = [0.0007872786903040108]
    beamline.sigma_y_coeffs = [1.9643343053823967]
    beamline.theta_y_coeffs = [0.0007911780133478402]
    beamline.epsilon_y_coeffs = [0.0024916149017600447]

    # NOTE: HBL means that the beam is coming from -x (90 degree rot around y)
    nSim = 20000  # particles to simulate per beam
    # rt_plan = ref_path / "RP1.2.752.243.1.1.20220406175810679.4500.52008_tagman.dcm"
    # beamset = BeamsetInfo(rt_plan)
    # G = float(beamset.beam_angles[0])

    ## ----  VBL Nozzle  ---
    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.rotation = Rotation.from_euler("x", -90, degrees=True).as_matrix()
    box.translation = [0.0, -1148 * mm, 0.0]
    box.material = "Vacuum"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box.name
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # Rashi
    rashi = sim.add_volume("Box", "rashi")
    rashi.mother = box.name
    rashi.size = [500 * mm, 500 * mm, 5 * mm]
    rashi.translation = [0.0, 0.0, 200 * mm]
    rashi.material = "G4_LUCITE"
    rashi.color = [1, 0, 1, 1]

    ## ----  HBL Nozzle  ---
    # FIXME : will change after volumes are refactored
    box_rot = sim.add_volume("Box", "box_rot")
    box_rot.configure_like(box)
    box_rot.rotation = Rotation.from_euler("y", -90, degrees=True).as_matrix()
    box_rot.translation = [1148.0, 0.0, 1000.0]

    nozzle_rot = sim.add_volume("Box", "nozzle_rot")
    nozzle_rot.configure_like(nozzle)
    nozzle_rot.mother = box_rot.name

    rashi_rot = sim.add_volume("Box", "rashi_rot")
    rashi_rot.configure_like(rashi)
    rashi_rot.mother = box_rot.name

    # -----------------------------------

    # target 1 VBL
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [324 * mm, 324 * mm, 324 * mm]
    phantom.translation = [0 * mm, 0.0, 0.0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # target 2 HBL
    phantom_rot = sim.add_volume("Box", "phantom_rot")
    phantom_rot.configure_like(phantom)
    phantom_rot.rotation = Rotation.from_euler("z", 90, degrees=True).as_matrix()
    phantom_rot.translation = [0.0, 0.0, 1000.0]

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "doseInXYZ")
    dose_actor.output_filename = "testTPSgantry.mhd"
    dose_actor.attached_to = phantom
    dose_actor.size = [162, 1620, 162]
    dose_actor.spacing = [2.0, 0.2, 2.0]
    dose_actor.hit_type = "random"
    dose_actor.dose.active = True

    dose_actor_rot = sim.add_actor("DoseActor", "doseInXYZ_rot")
    dose_actor_rot.configure_like(dose_actor)
    dose_actor_rot.attached_to = phantom_rot
    dose_actor_rot.output_filename = "testTPSganry_rot.mhd"

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add TPSources
    beam_data_dict = spots_info_from_txt(
        ref_path / "TreatmentPlan4Gate-1D_HBL_120.txt", "ion 6 12", beam_nr=1
    )

    tps_rot = sim.add_source("TreatmentPlanPBSource", "HBL")
    tps_rot.n = nSim
    tps_rot.end_time = 0.5 * s
    tps_rot.beam_model = beamline
    tps_rot.beam_data_dict = beam_data_dict
    tps_rot.beam_nr = 1
    tps_rot.particle = "ion 6 12"
    tps_rot.position.translation = [0.0, 0.0, 1000.0]
    tps_rot.sorted_spot_generation = True

    beam_data_dict_vbl = beam_data_dict.copy()
    beam_data_dict_vbl["gantry_angle"] = 0

    tps = sim.add_source("TreatmentPlanPBSource", "VBL")
    tps.n = nSim
    tps.start_time = 0.5 * s
    tps.beam_model = beamline
    tps.beam_data_dict = beam_data_dict_vbl
    tps.beam_nr = 1
    tps.particle = "ion 6 12"
    tps.sorted_spot_generation = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    ## ------ TESTS -------##

    # ABSOLUTE DOSE
    # ok = utility.assert_images(
    #     dose.output,
    #     dose_rot.output,
    #     stat,
    #     tolerance=50,
    #     ignore_value_data2=0,
    # )
    ok = True

    # read output and ref
    img_mhd_out = itk.imread(output_path / dose_actor_rot.edep.get_output_path())
    img_mhd_ref = itk.imread(output_path / dose_actor.edep.get_output_path())
    data = itk.GetArrayViewFromImage(img_mhd_out)
    data_ref = itk.GetArrayViewFromImage(img_mhd_ref)
    spacing = img_mhd_out.GetSpacing()

    # Range 80
    ok = (
        utility.compareRange(
            data, data_ref, data.shape, data_ref.shape, spacing, spacing
        )
        and ok
    )

    # 1D plots
    # fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    # gate.plot_img_axis(ax, img_mhd_out, "y profile", axis="y")
    # gate.plot_img_axis(ax, img_mhd_ref, "y ref", axis="y")
    # fig.savefig(output_path / "dose_profiles_water.png")
    # plt.show()

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.geometry.volumes import unite_volumes, subtract_volumes, intersect_volumes
import opengate_core as g4
from scipy.spatial.transform import Rotation

import opengate.tests.utility as tu


# the function called 'after_init' MUST be defined outside the main block
# It is used as "hook function" in the simulation and MUST take a simulation_engine object
# as input argument. No other arguments are accepted.
def after_init(simulation_engine):
    print("Checking solid ...")
    v = simulation_engine.volume_engine.get_volume("my_stuff").g4_logical_volume
    is_ok = v.GetName() == "my_stuff"
    tu.print_test(is_ok, f"Get volume {v.GetName()}")
    solid = v.GetSolid()
    pMin = g4.G4ThreeVector()
    pMax = g4.G4ThreeVector()
    solid.BoundingLimits(pMin, pMax)
    is_ok = list(pMin) == list([-50, -90, -100]) and is_ok
    tu.print_test(is_ok, f"pMin {pMin}")
    is_ok = list(pMax) == list([50, 60, 100]) and is_ok
    tu.print_test(is_ok, f"pMax {pMax}")
    if not is_ok:
        tu.test_ok(is_ok)


if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__)

    # global log level
    # create the simulation
    sim = gate.Simulation()
    print(f"Volumes types: {sim.volume_manager.dump_volume_types()}")

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.check_volumes_overlap = True
    sim.random_seed = 1236789

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # shortcuts for units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # first create the volumes to be used in the boolean operations
    b = gate.geometry.volumes.BoxVolume(name="box")
    b.size = [10 * cm, 10 * cm, 10 * cm]
    s = gate.geometry.volumes.SphereVolume(name="test_sph")
    s.rmax = 5 * cm
    t = gate.geometry.volumes.TubsVolume(name="t")
    t.rmin = 0
    t.rmax = 2 * cm
    t.dz = 15 * cm

    # boolean operations
    a1 = unite_volumes(b, s, translation=[0, 1 * cm, 5 * cm])
    a2 = subtract_volumes(a1, t, translation=[0, 1 * cm, 5 * cm])
    a3 = unite_volumes(a2, b, translation=[0, -1 * cm, -5 * cm])
    c = intersect_volumes(t, s, translation=[3 * cm, 0, 0])
    u = unite_volumes(a3, c, translation=[0, -7 * cm, -5 * cm], new_name="my_stuff")

    # Set user infos of this new volume
    u.translation = [5 * cm, 5 * cm, 5 * cm]
    u.rotation = Rotation.from_euler("x", 33, degrees=True).as_matrix()
    u.mother = "world"
    u.material = "G4_WATER"
    u.color = [0, 1, 0, 1]
    # add the new volume to the simulation
    sim.volume_manager.add_volume(u)

    # create a volume from a solid (not really useful)
    sim.volume_manager.add_volume(s)
    s.translation = [-5 * cm, -5 * cm, 1 - 5 * cm]
    s.mother = "world"
    s.material = "G4_WATER"
    s.color = [0, 1, 1, 1]

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "proton"
    source.energy.mono = 240 * MeV
    source.position.radius = 1 * cm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 5 * Bq

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # function to run after init

    # create G4 objects
    print(sim)

    # start simulation
    sim.user_hook_after_init = after_init
    sim.run(True)

    # print results at the end
    print(stats)

    tu.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test009_voxels", "test020")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    print(sim)

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.material = "G4_WATER"
    fake.color = [1, 0, 1, 1]
    fake.rotation = Rotation.from_euler("x", 2, degrees=True).as_matrix()

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.mother = "fake"
    patient.material = "G4_AIR"  # default material
    vm = gate.geometry.materials.read_voxel_materials(
        paths.gate_data / "patient-HU2mat-v1.txt"
    )
    vm[0][0] = -10000
    patient.voxel_materials = vm
    patient.dump_label_image = paths.output / "test020_labels.mhd"

    # activity
    activity = 100 * kBq

    # source 1
    source = sim.add_source("GenericSource", "source1")
    source.energy.mono = 150 * keV
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = 10 * mm
    source.position.translation = [0, 0, -15 * cm]
    source.activity = activity / sim.number_of_threads
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]

    # large cuts, no e- needed
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    sim.physics_manager.global_production_cuts.gamma = 700 * um
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * m
    sim.physics_manager.global_production_cuts.proton = 1 * m

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = paths.output / "test20-edep.mhd"
    dose.attached_to = "patient"
    dose.size = [100, 100, 100]
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.output_coordinate_system = "attached_to_image"
    dose.translation = [0 * mm, 0 * mm, 1 * mm]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulations
    sim.run()

    # print results at the end
    print(stats)
    print(dose)

    # tests
    stats_ref = utility.read_stat_file(paths.gate / "output" / "stat_profiling.txt")
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stats, stats_ref, 0.1)
    is_ok = is_ok and utility.assert_images(
        paths.gate / "output" / "output_profiling-Edep.mhd",
        dose.edep.get_output_path(),
        stats,
        tolerance=79,
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test019_linac_phsp_helpers as test019

if __name__ == "__main__":
    sim = test019.init_test019(3, "linacphsp_mt")
    test019.run_test019(sim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test064")
    output_path = paths.output

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 2
    sim.random_seed = 987654321
    sim.output_dir = output_path

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    gcm3 = gate.g4_units["g/cm3"]

    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"

    # nb of particles
    n = 2000

    # source1
    source1 = sim.add_source("GenericSource", "photon_source")
    source1.particle = "gamma"
    source1.position.type = "box"
    source1.position.size = [1 * cm, 1 * cm, 1 * cm]
    source1.position.translation = [10 * cm, 0 * cm, 3 * cm]
    source1.direction.type = "momentum"
    source1.direction.momentum = [0, 0, -1]
    source1.energy.type = "mono"
    source1.energy.mono = 1 * MeV
    source1.n = n / sim.number_of_threads

    # source2
    source2 = sim.add_source("GenericSource", "photon_source_2")
    source2.particle = "gamma"
    source2.position.type = "point"
    source2.position.size = [1 * cm, 1 * cm, 1 * cm]
    source2.position.translation = [-10 * cm, 0 * cm, -3 * cm]
    source2.direction.type = "momentum"
    source2.direction.momentum = [0, 0, 1]
    source2.energy.type = "mono"
    source2.energy.mono = 1 * MeV
    source2.n = n / sim.number_of_threads

    # stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # add phase space plan
    phsp_plane = sim.add_volume("Box", "phase_space_plane")
    phsp_plane.mother = world
    phsp_plane.material = "G4_Galactic"
    phsp_plane.size = [1 * m, 1 * m, 1 * nm]
    phsp_plane.translation = [0, 0, 0]
    phsp_plane.color = [1, 0, 0, 1]  # red

    # PhaseSpace Actor
    phsp = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp.attached_to = phsp_plane
    phsp.attributes = ["EventID"]
    phsp.output_filename = "test064.root"

    # kill actor volume
    kill_plane = sim.add_volume("Box", "Kill_plane")
    kill_plane.mother = world
    kill_plane.material = "G4_Galactic"
    kill_plane.size = [10 * cm, 10 * cm, 1 * nm]
    kill_plane.translation = [-10 * cm, 0, -2 * cm]
    kill_plane.color = [0, 1, 0, 1]  # green

    # kill actor
    kill_actor = sim.add_actor("KillActor", "KillAct")
    kill_actor.attached_to = kill_plane

    # Physic list and cuts
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm

    # go !
    sim.run()

    # print results
    print(stats)

    f_phsp = uproot.open(phsp.get_output_path())
    arr = f_phsp["PhaseSpace"].arrays()
    print("Number of detected events :", len(arr))
    print("Number of expected events :", n)
    # EventID = arr[0]

    # Nb of kill
    nk = kill_actor.number_of_killed_particles
    print(f"Number of kills = {nk}")

    is_ok = len(arr) == n
    is_ok = (nk == n) and is_ok
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import os
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", "test044_pbs_weight"
    )
    output_path = paths.output

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123654789
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    mrad = gate.g4_units.mrad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    ## FIRST SOURCE + DETECTOR ##
    # waterbox
    # translation and rotation like in the Gate macro
    waterbox1 = sim.add_volume("Box", "waterbox1")
    waterbox1.size = [10 * cm, 10 * cm, 105 * cm]
    waterbox1.translation = [0 * cm, 0 * cm, 52.5 * cm]
    waterbox1.material = "Vacuum"
    waterbox1.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom = sim.add_volume("Box", "phantom_a_1")
    phantom.mother = "waterbox1"
    phantom.size = [100 * mm, 100 * mm, 50 * mm]
    phantom.translation = [0 * mm, 0 * mm, -500 * mm]
    phantom.rotation = m
    phantom.material = "G4_AIR"
    phantom.color = [1, 0, 1, 1]

    # default source for tests (from test42)
    source = sim.add_source("IonPencilBeamSource", "mysource1")
    source.attached_to = "waterbox1"
    source.energy.mono = 60 * MeV
    source.particle = "proton"
    source.position.type = "disc"  # pos = Beam, shape = circle + sigma
    source.position.translation = [0 * mm, 0 * mm, -52.5 * cm]
    # rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 20000
    source.weight = 1
    source.direction.partPhSp_x = [
        2.3335754 * mm,
        2.3335754 * mrad,
        0.00078728 * mm * mrad,
        0,
    ]
    source.direction.partPhSp_y = [
        1.96433431 * mm,
        0.00079118 * mrad,
        0.00249161 * mm * mrad,
        0,
    ]

    # add dose actor
    dose = sim.add_actor("DoseActor", "doseInYZ_1")
    filename = "phantom_a_1.mhd"
    dose.output_filename = output_path / filename
    dose.attached_to = "phantom_a_1"
    dose.size = [250, 250, 1]
    dose.spacing = [0.4, 0.4, 2]
    dose.hit_type = "random"

    ## SECOND SOURCE + DETECTOR ##
    # waterbox
    # translation and rotation like in the Gate macro
    waterbox2 = sim.add_volume("Box", "waterbox2")
    waterbox2.size = [10 * cm, 10 * cm, 105 * cm]
    waterbox2.translation = [30 * cm, 0 * cm, 52.5 * cm]
    waterbox2.material = "Vacuum"
    waterbox2.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom2 = sim.add_volume("Box", "phantom_a_2")
    phantom2.mother = "waterbox2"
    phantom2.size = [100 * mm, 100 * mm, 50 * mm]
    phantom2.translation = [0 * mm, 0 * mm, -500 * mm]
    phantom2.rotation = m
    phantom2.material = "G4_AIR"
    phantom2.color = [1, 0, 1, 1]

    # default source for tests (from test42)
    source2 = sim.add_source("IonPencilBeamSource", "mysource2")
    source2.attached_to = "waterbox2"
    source2.energy.mono = 60 * MeV
    source2.particle = "proton"
    source2.position.type = "disc"  # pos = Beam, shape = circle + sigma
    source2.position.translation = [0 * mm, 0 * mm, -52.5 * cm]
    # rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
    source2.direction.type = "momentum"
    source2.direction.momentum = [0, 0, 1]
    source2.n = 20000
    source2.weight = 2
    source2.direction.partPhSp_x = [
        2.3335754 * mm,
        2.3335754 * mrad,
        0.00078728 * mm * mrad,
        0,
    ]
    source2.direction.partPhSp_y = [
        1.96433431 * mm,
        0.00079118 * mrad,
        0.00249161 * mm * mrad,
        0,
    ]

    # add dose actor
    dose2 = sim.add_actor("DoseActor", "doseInYZ_2")
    filename = "phantom_a_2.mhd"
    dose2.output_filename = filename
    dose2.attached_to = "phantom_a_2"
    dose2.size = [250, 250, 1]
    dose2.spacing = [0.4, 0.4, 2]
    dose2.hit_type = "random"

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.global_production_cuts.all = 1000 * km

    print(sim.source_manager.dump_sources())

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # print results at the end
    print(stat)

    # ----------------------------------------------------------------------------------------------------------------
    # tests

    # energy deposition: we expect the edep from source two
    # to be double the one of source one

    print("\nDifference for EDEP")
    mhd_1 = sim.get_actor("doseInYZ_1").get_output_path("edep")
    mhd_2 = sim.get_actor("doseInYZ_2").get_output_path("edep")
    test = True
    # test = utility.assert_images(
    #     output_path / mhd_1,
    #     output_path / mhd_2,
    #     stat,
    #     axis="x",
    #     tolerance=50,
    #     ignore_value_data2=0,
    # )
    fig1 = utility.create_2D_Edep_colorMap(output_path / mhd_1, show=False)
    fig2 = utility.create_2D_Edep_colorMap(output_path / mhd_2, show=False)

    # Total Edep
    is_ok = (
        utility.test_weights(
            source2.weight / source.weight,
            output_path / mhd_1,
            output_path / mhd_2,
            thresh=0.2,
        )
        and test
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot
import numpy as np

#########################################################################################
# Simulations configuration that may be relevant to change
#########################################################################################
# Number of back-to-back to generate
nbEvents = 100

#########################################################################################
# Constants
#########################################################################################
# Units
MeV = gate.g4_units.MeV
keV = gate.g4_units.keV
Bq = gate.g4_units.Bq
deg = gate.g4_units.deg
mm = gate.g4_units.mm
m = gate.g4_units.m
cm = gate.g4_units.cm


#########################################################################################
# Methods used in this test
#########################################################################################
def test_back_to_back(_path_to_root_file, _nb_b2b):
    # get data from root file
    b2b_root = uproot.open(_path_to_root_file)["phsp;1"].arrays(library="numpy")

    # Assuming that trackId is relative to a eventId, it provides use a way to check
    # that everything start with two gammas.
    # It does not take into account 3+ gamma cases, but I do not know how to do better
    b2b_eventId_adamEve = b2b_root["TrackID"] <= 2

    b2b_emissionEnergy = b2b_root["EventKineticEnergy"][b2b_eventId_adamEve]

    b2b_particleName = b2b_root["ParticleName"][b2b_eventId_adamEve]

    b2b_dir = (
        np.stack(
            (
                b2b_root["PreDirection_X"],
                b2b_root["PreDirection_Y"],
                b2b_root["PreDirection_Z"],
            )
        ).T
    )[b2b_eventId_adamEve]
    # For easier manipulation
    b2b_dir = b2b_dir.reshape((_nb_b2b, 2, 3))

    # Note: We assumes that thing that are not specific to back-to-back were already
    # tested, for example the number of particle
    # FIXME Would be better to check fAccolinearityFlag but might not be accessible?
    is_b2b = source.particle == "back_to_back"
    is_monoEnergy = source.energy.type == "mono"
    is_def511kev = source.energy.mono == 511 * keV
    is_defNoAcolin = source.direction.accolinearity_flag == False
    is_allB2b511keV = np.all(np.isclose(b2b_emissionEnergy, 0.511, atol=10**-3))
    is_allEmissionGamme = np.all(b2b_particleName == "gamma")
    is_b2bColin = np.all(
        np.isclose(
            np.sum(b2b_dir[:, 0, :] * b2b_dir[:, 1, :], axis=-1), -1.0, atol=10**-3
        )
    )
    # FIXME: Confirm it stay in the sphere?

    print(f"is_b2: {is_b2b}")
    print(f"is_monoEnergy: {is_monoEnergy}")
    print(f"is_def511kev: {is_def511kev}")
    print(f"is_defNoAcolin: {is_defNoAcolin}")
    print(f"is_allB2b511keV: {is_allB2b511keV}")
    print(f"is_allEmissionGamme: {is_allEmissionGamme}")
    print(f"is_b2bColin: {is_b2bColin}")

    return (
        is_b2b
        & is_monoEnergy
        & is_def511kev
        & is_defNoAcolin
        & is_allB2b511keV
        & is_allEmissionGamme
        & is_b2bColin
    )


#########################################################################################
# Main : We use this to launch the test
#########################################################################################
if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test072")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.number_of_threads = 1
    sim.random_seed = 123456
    sim.output_dir = paths.output

    # set the world size like in the Gate macro
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # test sources
    source = sim.add_source("GenericSource", "b2b")
    source.particle = "back_to_back"
    source.n = nbEvents
    source.position.type = "sphere"
    source.position.radius = 5 * mm
    source.direction.type = "iso"
    source.direction.accolinearity_flag = False
    # note : source.energy is ignored (always 511 keV)
    # FIXME : do another test with accolinearity_flag set to True

    # actors
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")

    # store phsp
    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp_actor.attributes = [
        "EventID",
        "TrackID",
        "EventPosition",
        "EventDirection",
        "Direction",
        "EventKineticEnergy",
        "KineticEnergy",
        "ParticleName",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "PDGCode",
        "PostPosition",
        "PreDirection",
    ]
    phsp_actor.steps_to_store = "first"
    phsp_actor.output_filename = "b2b.root"

    # verbose
    # sim.g4_verbose = True
    # sim.g4_commands_after_init.append("/tracking/verbose 2")
    # sim.g4_commands_after_init.append("/run/verbose 2")
    # sim.g4_commands_after_init.append("/event/verbose 2")
    # sim.g4_commands_after_init.append("/tracking/verbose 1")

    # start simulation
    sim.run()

    # test
    is_ok = test_back_to_back(phsp_actor.get_output_path(), nbEvents)

    # this is the end, my friend
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 2 09:48:30 2025

@author: fava
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.utility import g4_units
import opengate.tests.utility as utility


def simuation_IDD(test_material):
    paths = utility.get_default_test_paths(__file__, "gate_test044_pbs")

    # units
    eV = g4_units.eV
    MeV = g4_units.MeV
    cm = g4_units.cm
    mm = g4_units.mm

    # create simulation object
    sim = gate.Simulation()
    # sim.number_of_threads = 4
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 10 * cm, 10 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    phantom_off = sim.add_volume("Box", "phantom_off")
    phantom_off.mother = phantom.name
    phantom_off.size = [100 * mm, 60 * mm, 60 * mm]
    phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
    phantom_off.material = test_material
    phantom_off.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 1424 * MeV
    source.particle = "ion 6 12"

    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.radius = 4 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    # print(dir(source.energy))
    source.n = 1e3
    # source.activity = 100 * kBq

    size = [500, 1, 1]
    spacing = [0.2 * mm, 60.0 * mm, 60.0 * mm]

    doseActorName_IDD_d = "IDD"
    doseIDD = sim.add_actor("DoseActor", doseActorName_IDD_d)
    doseIDD.output_filename = paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")
    #    print(f'actor: {paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")}')
    doseIDD.attached_to = phantom_off.name
    doseIDD.size = size
    doseIDD.spacing = spacing
    doseIDD.hit_type = "random"
    doseIDD.dose.active = True

    return sim, doseIDD


if __name__ == "__main__":
    # units
    eV = g4_units.eV
    cm = g4_units.cm

    # overrides for material ionisation potential
    test_material = "Water"
    materials_Ival_dict = {test_material: 30 * eV}

    sim, dose = simuation_IDD(test_material)

    # dump ionisation value before modification
    mat = sim.volume_manager.find_or_build_material(test_material)
    ionisation = mat.GetIonisation().GetMeanExcitationEnergy()
    print(f"Default ionisation potential of {test_material}: {ionisation} eV")

    # run simulation with default value for material
    sim.run(True)

    # get idd
    idd_pre = np.flip(np.asarray(dose.dose.image), axis=2)
    x1, d1_pre = utility.get_1D_profile(
        idd_pre, np.flip(dose.size), dose.spacing, axis="x"
    )
    r80_pre, _ = utility.getRange(x1, d1_pre)
    plt.plot(x1, d1_pre)
    print(f"Range in water: {r80_pre} mm")

    # run simulation again, but with modified ionisation potential
    sim, dose = simuation_IDD(test_material)
    sim.physics_manager.material_ionisation_potential = materials_Ival_dict

    # run simulation with new value
    sim.run(True)

    # dump ionisation value after modification
    ionisation = mat.GetIonisation().GetMeanExcitationEnergy()
    print(f"New ionisation potential of {test_material}: {ionisation} eV")

    # get idd
    idd_after = np.flip(np.asarray(dose.dose.image), axis=2)
    x1, d1_after = utility.get_1D_profile(
        idd_after, np.flip(dose.size), dose.spacing, axis="x"
    )
    r80_after, _ = utility.getRange(x1, d1_after)
    plt.plot(x1, d1_after)
    print(f"Range in water: {r80_after} mm")

    ok = (r80_pre - r80_after) > dose.spacing[0]
    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test030"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 983456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    um = gate.g4_units.um
    nm = gate.g4_units.nm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = fake
    waterbox.size = [20 * cm, 20 * cm, 20 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.set_production_cut("world", "all", 700 * um)

    # default source for tests
    # the source is fixed at the center, only the volume will move
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 150 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 30000 * Bq

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test030-dyn.mhd"
    dose.attached_to = waterbox
    dose.size = [99, 99, 99]
    mm = gate.g4_units.mm
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.edep.keep_data_per_run = True
    dose.edep.auto_merge = True
    dose.edep_uncertainty.active = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # motion
    n = 3
    interval_length = 1 * sec / n
    sim.run_timing_intervals = [
        (i * interval_length, (i + 1) * interval_length) for i in range(n)
    ]
    gantry_angles_deg = [i * 20 for i in range(n)]
    (
        dynamic_translations,
        dynamic_rotations,
    ) = gate.geometry.utility.get_transform_orbiting(
        initial_position=fake.translation, axis="Y", angle_deg=gantry_angles_deg
    )
    fake.add_dynamic_parametrisation(
        translation=dynamic_translations, rotation=dynamic_rotations
    )

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "stats030.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.11)

    print()
    gate.exception.warning("Difference for EDEP")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test030-edep.mhd",
            dose.edep.get_output_path(),
            stats,
            tolerance=30,
            sum_tolerance=6,
            ignore_value_data2=0,
        )
        and is_ok
    )

    print("\nDifference for uncertainty")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test030-edep_uncertainty.mhd",
            dose.edep_uncertainty.get_output_path(),
            stats,
            tolerance=15,
            ignore_value_data2=0,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.sources.utility import set_source_energy_spectrum
from opengate.exception import warning
from opengate.tests import utility
import opengate.contrib.spect.siemens_intevo as intevo
import opengate.contrib.spect.ge_discovery_nm670 as nm670
from opengate.contrib.spect.spect_helpers import get_default_energy_windows
from scipy.spatial.transform import Rotation
import itk
import numpy as np


def create_sim_tests(sim, threads=1, digitizer=1, debug=False):
    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = threads
    sim.visu = False

    # activity
    activity = 7000000 * Bq / sim.number_of_threads

    # world size
    world = sim.world
    world.size = [2.2 * m, 3.2 * m, 2 * m]
    world.material = "G4_AIR"

    # spect head
    head, _, crystal = intevo.add_spect_head(
        sim, "spect", collimator_type="melp", debug=debug
    )
    head.translation = [0, 0, -280 * mm]
    head.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()

    # spect digitizer
    if digitizer == 1:
        channels = get_default_energy_windows("lu177", spectrum_channel=True)
        intevo.add_digitizer(sim, crystal.name, channels=channels)
    if digitizer == 2:
        channels = get_default_energy_windows("lu177", spectrum_channel=True)
        intevo.add_digitizer(sim, crystal.name, channels=channels)
        # change parameters to add a fake blurring
        keV = gate.g4_units.keV
        MeV = gate.g4_units.MeV
        eb = sim.actor_manager.find_actor_by_type("DigitizerBlurringActor")
        sb = sim.actor_manager.find_actor_by_type("DigitizerSpatialBlurringActor")
        eb.blur_attribute = "TotalEnergyDeposit"
        eb.blur_method = "Linear"
        eb.blur_resolution = 0.13
        eb.blur_reference_value = 80 * keV
        eb.blur_slope = -0.09 * 1 / MeV
        sb.blur_fwhm = 10 * mm

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.all = 100 * mm
    sim.physics_manager.set_production_cut("spect", "all", 0.1 * mm)

    # sources
    # sim.running_verbose_level = gate.EVENT
    s1 = sim.add_source("GenericSource", "s1")
    s1.particle = "gamma"
    s1.position.type = "sphere"
    s1.position.radius = 30 * mm
    s1.position.translation = [0, 0, 0]
    set_source_energy_spectrum(s1, "Lu177", "radar")
    s1.direction.type = "iso"
    s1.activity = activity

    s2 = sim.add_source("GenericSource", "s2")
    s2.particle = "gamma"
    s2.position.type = "sphere"
    s2.position.radius = 60 * mm
    s2.position.translation = [0, 200 * mm, 0]
    set_source_energy_spectrum(s2, "Lu177", "radar")
    s2.direction.type = "iso"
    s2.activity = activity

    s3 = sim.add_source("GenericSource", "s3")
    s3.particle = "gamma"
    s3.position.type = "sphere"
    s3.position.radius = 25 * mm
    s3.position.translation = [100, 0, 0 * mm]
    set_source_energy_spectrum(s3, "Lu177", "radar")
    s3.direction.type = "iso"
    s3.activity = activity

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True
    s.output_filename = "stats.txt"

    # timing
    sec = gate.g4_units.second
    stop = 1 * sec
    if sim.visu:
        stop = 0.001 * sec
    sim.run_timing_intervals = [[0, stop]]


def compare_stats(sim, filename):
    # compare stats
    warning("Compare stats")
    stats = sim.get_actor("stats")
    # force nb of thread to 1
    stats_ref = utility.read_stat_file(filename)
    stats.counts.runs = stats_ref.counts.runs
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.01)
    return is_ok


def compare_root_hits(crystal, sim, root_filename, path, n=1):
    # compare root
    print()
    warning("Compare hits")
    gate_file = root_filename
    hc = sim.actor_manager.find_actor_by_type(f"DigitizerHitsCollectionActor")
    hc_file = hc.get_output_path()
    checked_keys = [
        {"k1": "posX", "k2": "PostPosition_X", "tol": 3.2, "scaling": 1},
        {"k1": "posY", "k2": "PostPosition_Y", "tol": 7, "scaling": 1},
        {"k1": "posZ", "k2": "PostPosition_Z", "tol": 0.21, "scaling": 1},
        {"k1": "edep", "k2": "TotalEnergyDeposit", "tol": 0.005, "scaling": 1},
        {"k1": "time", "k2": "GlobalTime", "tol": 0.05, "scaling": 1e-9},
    ]
    is_ok = utility.compare_root2(
        gate_file,
        hc_file,
        "Hits",
        f"{crystal.name}_hits",
        checked_keys,
        path / f"test073_test_{n}_hits.png",
        n_tol=16,
    )
    return is_ok


def compare_root_singles(crystal, sim, root_filename, path, sname, n=1):
    # Compare root files
    print()
    warning("Compare singles")
    # hc_file = sim.get_actor(f"Singles_{crystal.name}").get_output_path()
    hc = sim.actor_manager.find_actor_by_type(f"DigitizerAdderActor")
    hc_file = hc.get_output_path()
    checked_keys = [
        {"k1": "globalPosX", "k2": "PostPosition_X", "tol": 4.5, "scaling": 1},
        {"k1": "globalPosY", "k2": "PostPosition_Y", "tol": 9.0, "scaling": 1},
        {"k1": "globalPosZ", "k2": "PostPosition_Z", "tol": 0.29, "scaling": 1},
        {"k1": "energy", "k2": "TotalEnergyDeposit", "tol": 0.0045, "scaling": 1},
    ]
    is_ok = utility.compare_root2(
        root_filename,
        hc_file,
        "Singles",
        sname,
        checked_keys,
        f"{path}/test050_test_{n}_singles.png",
        n_tol=9,
    )
    return is_ok


def compare_proj_images(crystal, sim, stats, image_filename, path, n=1):
    # compare images with Gate
    print()
    print("Compare images (old spacing/origin)")
    # read the image and force change the offset to be similar to old Gate
    proj = sim.actor_manager.find_actor_by_type(f"DigitizerProjectionActor")
    fr = image_filename
    f1 = path / f"projections_test{n}_counts.mhd"
    f2 = path / f"projections_test{n}_offset.mhd"
    img = itk.imread(f1)
    spacing = np.array(proj.counts.image.GetSpacing())
    origin = spacing / 2.0
    origin[2] = 0.5
    spacing[2] = 1
    img.SetSpacing(spacing)
    img.SetOrigin(origin)
    itk.imwrite(img, f2)

    is_ok = utility.assert_images(
        fr,
        f2,
        stats,
        tolerance=69,
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,
        axis="y",
        fig_name=path / f"test073_test_{n}.png",
        sum_tolerance=6,
    )
    return is_ok


def test073_setup_sim(sim, spect_type, collimator_type):
    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 4
    # sim.visu = True
    sim.visu_type = "qt"
    # sim.random_seed = 321654987

    # world size
    world = sim.world
    world.size = [2.2 * m, 3.2 * m, 2 * m]
    world.material = "G4_AIR"

    # spect ?
    if spect_type == "intevo":
        head, _, _ = intevo.add_spect_head(
            sim, "spect", collimator_type=collimator_type, debug=sim.visu
        )
        head.translation = [0, 0, -280 * mm]
        head.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    if spect_type == "discovery":
        head, _, _ = nm670.add_spect_head(
            sim, "spect", collimator_type=collimator_type, debug=sim.visu
        )
        head.rotation = Rotation.from_euler("z", 9, degrees=True).as_matrix()
        head.translation = [0, 0, -280 * mm]

    # source
    source = sim.add_source("GenericSource", "source")
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = 30 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "iso"
    source.direction.acceptance_angle.volumes = [head.name]
    source.direction.acceptance_angle.intersection_flag = True

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.all = 100 * mm
    sim.physics_manager.set_production_cut("spect", "all", 0.1 * mm)

    return head, stats, source


def compare_root_spectrum(ref_output, output, png_filename):
    # compare root
    print()
    warning("Compare spectrum")
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 1.6, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 1.7, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.3, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.003,
            "scaling": 1,
        },
        {"k1": "GlobalTime", "k2": "GlobalTime", "tol": 1.6e7, "scaling": 1},
    ]
    is_ok = utility.compare_root2(
        ref_output,
        output,
        "spectrum",
        "spectrum",
        checked_keys,
        png_filename,
        n_tol=25,
    )
    return is_ok


def compare_root_spectrum2(ref_output, output, png_filename):
    # compare root
    print()
    warning("Compare spectrum")
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 2.0, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 2.0, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.4, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.005,
            "scaling": 1,
        },
        {"k1": "GlobalTime", "k2": "GlobalTime", "tol": 1.6e7, "scaling": 1},
    ]
    is_ok = utility.compare_root2(
        ref_output,
        output,
        "spectrum",
        "spectrum",
        checked_keys,
        png_filename,
        n_tol=16,
    )
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate import g4_units
import opengate.contrib.pet.philipsvereos as pet_vereos
from opengate.geometry.utility import get_grid_repetition

if __name__ == "__main__":
    # initialize the simulation
    sim = gate.Simulation()
    sim.number_of_threads = 8

    # world
    m = g4_units.m
    sim.world.size = [3 * m, 3 * m, 3 * m]
    sim.world.material = "G4_AIR"

    # waterbox
    cm = g4_units.cm
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [10 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # insert a TEP
    pet = pet_vereos.add_pet(sim, "pet")

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 1 * m)
    sim.physics_manager.set_production_cut("waterbox", "gamma", 0.1 * cm)

    # source
    MBq = 1e6 * g4_units.Bq
    source = sim.add_source("GenericSource", "source")
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.particle = "e+"
    source.energy.type = "F18"
    source.direction.type = "iso"
    source.activity = 300 * MBq

    # repeat crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.size = [1 * cm, 1 * cm, 1 * cm]
    crystal.translation = get_grid_repetition([2, 3, 4], [0.5 * cm, 0.5 * cm, 0.5 * cm])

    # go
    sim.run()

    # get results
    output = sim.output



import os
import numpy as np
import re
from box import Box
import itk

import opengate_core as g4
from ..utility import fatal, g4_units, g4_best_unit
from ..definitions import elements_name_symbol


def read_voxel_materials(filename, def_mat="G4_AIR"):
    p = os.path.abspath(filename)
    current = 0
    materials = []
    with open(p, "r") as f:
        for line in f:
            for word in line.split():
                if word[0] == "#":
                    break
                if current == 0:
                    start = float(word)
                    current = 1
                else:
                    if current == 1:
                        stop = float(word)
                        current = 2
                    else:
                        if current == 2:
                            mat = word
                            current = 0
                            materials.append([start, stop, mat])

    # sort according to starting interval
    materials = sorted(materials)

    # consider all values
    pix_mat = []
    previous = materials[0][0]
    for m in materials:
        if previous and previous > m[0]:
            fatal(
                f"Error while reading {filename}\n"
                f"Intervals are not disjoint: {previous} {m}"
            )
        if m[0] > m[1]:
            fatal(f"Error while reading {filename}\n" f"Wrong interval {m}")
        if not previous or previous == m[0]:
            pix_mat.append([previous, m[1], m[2]])
            previous = m[1]
        else:
            pix_mat.append([previous, m[0], def_mat])
            pix_mat.append([previous, m[1], m[2]])
            previous = m[1]

    return pix_mat


def HU_read_materials_table(file_mat):
    p = os.path.abspath(file_mat)
    elements = ["HU"]
    materials = []
    current_section = None
    current_material = None
    with open(p, "r") as f:
        for line in f:
            i = 0
            for word in line.split():
                if word[0] == "#":
                    break
                if word == "[Elements]":
                    current_section = "element"
                    break
                if word == "[/Elements]":
                    elements.append("name")
                    current_section = "table"
                    break
                if current_section is None:
                    break
                if current_section == "element":
                    elements.append(word)
                if current_section == "table":
                    if current_material is None:
                        current_material = {}
                    if i == 0:
                        current_material[elements[i]] = int(word)
                    else:
                        if i == len(elements) - 1:
                            current_material[elements[i]] = word
                        else:
                            current_material[elements[i]] = float(word)
                i += 1
            if current_material:
                materials.append(current_material)
            current_material = None
    return materials, elements


def HU_read_density_table(file_density):
    p = os.path.abspath(file_density)
    densities = []
    with open(p, "r") as f:
        for line in f:
            words = line.split()
            if len(words) < 1:
                continue
            if words[0][0] == "#":
                continue
            d = {"HU": float(words[0]), "density": float(words[1])}
            densities.append(d)
    return densities


def HU_linear_interpolate_densities(hu, densities):
    i = 0
    n = len(densities)
    while i < n and hu > densities[i]["HU"]:
        i = i + 1
    i = i - 1
    if i < 0:
        return densities[0]["density"]
    if i >= n - 1:
        return densities[n - 1]["density"]
    v = ((hu - densities[i]["HU"]) / (densities[i + 1]["HU"] - densities[i]["HU"])) * (
        densities[i + 1]["density"] - densities[i]["density"]
    ) + densities[i]["density"]
    return v


def HU_find_max_density_difference(hu_min, hu_max, d_min, d_max, densities):
    n = len(densities)
    i = 0
    while i < n and hu_min > densities[i]["HU"]:
        i = i + 1
    j = 0
    while j < n and hu_max > densities[j]["HU"]:
        j = j + 1
    j = j - 1
    for x in range(i, j, 1):
        if densities[x]["density"] < d_min:
            d_min = densities[x]["density"]
        if densities[x]["density"] > d_max:
            d_max = densities[x]["density"]
    return d_max - d_min


def HounsfieldUnit_to_material(simulation, density_tolerance, file_mat, file_density):
    """
    Same function than in GateHounsfieldToMaterialsBuilder class.
    Probably far from optimal, put we keep the compatibility
    """

    materials, elements = HU_read_materials_table(file_mat)
    densities = HU_read_density_table(file_density)
    voxel_materials = []
    created_materials = []
    gcm3 = g4_units.g_cm3

    elems = elements[1 : len(elements) - 1]
    elems_symbol = [elements_name_symbol[x] for x in elems]

    i = 0
    num = 0
    last_i = len(materials) - 1
    for mat in materials:
        # get HU interval
        hu_min = mat["HU"]
        if i == last_i:
            hu_max = hu_min + 1
        else:
            hu_max = materials[i + 1]["HU"]

        # check hu min max
        if hu_max <= hu_min:
            fatal(f"Error, HU interval not valid: {mat}")

        # get densities interval
        dmin = HU_linear_interpolate_densities(hu_min, densities)
        dmax = HU_linear_interpolate_densities(hu_max, densities)
        ddiff = HU_find_max_density_difference(hu_min, hu_max, dmin, dmax, densities)

        # nb of bins
        n = max(1, ddiff * gcm3 / density_tolerance)
        # n_naive = max(1, (dmax - dmin) * gcm3 / density_tolerance)

        # check if AIR
        if "Air" in mat["name"] or "AIR" in mat["name"]:
            n = 1

        # HU interval according to tolerance
        htol = (hu_max - hu_min) / n

        # like in Gate
        if n != 1:
            n = int(np.ceil(n))

        # loop on density interval
        for j in range(n):
            h1 = hu_min + j * htol
            h2 = min(hu_min + (j + 1) * htol, hu_max)
            d = HU_linear_interpolate_densities(h1 + (h2 - h1) / 2.0, densities)
            # create a new material with the interpolated density
            weights = [mat[x] for x in elems]
            weights_nz = []
            elems_symbol_nz = []
            # remove the weight equal to zero
            sum_of_weights = 0
            for a, e in zip(weights, elems_symbol):
                if a > 0:
                    weights_nz.append(a)
                    elems_symbol_nz.append(e)
                    sum_of_weights += a
            # normalise weight
            for k in range(len(weights_nz)):
                weights_nz[k] = weights_nz[k] / sum_of_weights
            # define a new material (will be created later at MaterialDatabase initialize)
            name = f'{mat["name"]}_{num}'
            simulation.volume_manager.material_database.add_material_weights(
                name, elems_symbol_nz, weights_nz, d * gcm3
            )
            # get the final correspondence
            c = [h1, h2, name]
            voxel_materials.append(c)
            created_materials.append(name)
            num = num + 1
        #
        i = i + 1
    return voxel_materials, created_materials


def dump_material_like_Gate(mat):
    s = f'{mat.GetName()}: d={g4_best_unit(mat.GetDensity(), "Volumic Mass")}; n={mat.GetNumberOfElements()}\n'
    i = 0
    for elem in mat.GetElementVector():
        s += f"+el: name={elem.GetName()}; f={mat.GetElementFraction(i)}\n"
        i += 1
    s += "\n"
    return s


def assert_same_material(m1, m2):
    if m1.name != m2.name:
        return False
    if np.fabs(m1.density - m2.density) / m1.density > 1e-2:
        print("Error while comparing materials", m1, m2)
        print(np.fabs(m1.density - m2.density) / m1.density)
        print(m1)
        print(m2)
        return False
    for e1 in m1.components:
        e2 = m2.components[elements_name_symbol[e1]]
        e1 = m1.components[e1]
        if elements_name_symbol[e1.name] != e2.name:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False
        if e1.n != e2.n:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False
        if np.fabs(e1.f - e2.f) / e1.f > 1e-2:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False

    return True


def read_next_line(f):
    line = f.readline()
    return line.strip().replace("\t", " ")


def read_tag(s, tag):
    w = s.split("=")
    if w[0].strip() != tag:
        return None
    value = w[1].strip()
    return value


def read_tag_with_unit(s, tag):
    w = s.split("=")
    if w[0].strip() != tag:
        return None
    w = w[1].split()
    value = float(w[0])
    u = g4_units[w[1].strip()]
    return value * u


def create_density_img(img_volume, material_database):
    """


    Parameters
    ----------
    img_volume : ImageVolume
        opengate ImageVolume class instance
    material_database : MaterialDatabase
        simulation.volume_manager.material_database

    Returns
    -------
    rho : itk.Image
        Image of the same size and resolution of the ct. The voxel value is the density of the voxel converted to g/cm3.

    """
    img_volume.load_input_image()
    act = itk.GetArrayFromImage(img_volume.itk_image)
    arho = np.zeros(act.shape, dtype=np.float32)

    for hu0, hu1, mat_name in img_volume.voxel_materials:
        if mat_name not in material_database.g4_materials:
            material_database.FindOrBuildMaterial(mat_name)
        arho[(act >= hu0) * (act < hu1)] = material_database.g4_materials[
            mat_name
        ].GetDensity()

    arho *= g4_units.cm3 / g4_units.g
    rho = itk.GetImageFromArray(arho)
    rho.CopyInformation(img_volume.itk_image)

    return rho


def create_mass_img(ct_itk, hu_density_file, overrides=dict()):
    """


    Parameters
    ----------
    ct_itk :itk.Image
        ct image
    hu_density_file : str
        filepath of the HU to density table
    overrides : dict, optional
        Dict where keys are HU to be overwritten and values
        are density values. The default is dict().

    Returns
    -------
    mass : itk.Image
        image of the same size and resolution of the ct. The voxel value is the mass of the voxel.
        Mass is returned in grams.

    """
    hlut = HU_read_density_table(hu_density_file)
    act = itk.GetArrayFromImage(ct_itk)
    amass = np.zeros(act.shape, dtype=np.float32)
    done = np.zeros(act.shape, dtype=bool)

    m = act < hlut[0]["HU"]
    amass[m] = hlut[0]["density"]
    done |= m
    m = act >= hlut[-1]["HU"]
    amass[m] = hlut[-1]["density"]
    done |= m

    # interpolate for intermediate values
    for hlut0, hlut1 in zip(hlut[:-1], hlut[1:]):
        hu0, rho0 = hlut0["HU"], hlut0["density"]
        hu1, rho1 = hlut1["HU"], hlut1["density"]
        m = (act >= hu0) * (act < hu1)
        assert not (m * done).any(), "programming error"
        amass[m] = rho0
        amass[m] += (act[m] - hu0) * (rho1 - rho0) / (hu1 - hu0)
        done |= m
    assert done.all(), "programming error"

    #  override density for specific HU values
    for hu, rho in overrides.items():
        assert hu == int(hu), "overrides must be given for integer HU values"
        assert rho >= 0, "override density values must be non-negative"
        m = act == hu
        amass[m] = rho
        done |= m

    spacing = ct_itk.GetSpacing()
    voxel_vol = (
        spacing[0] * spacing[1] * spacing[2] * 1e-3
    )  # density in g/cm3 -> spacing in mm
    amass *= voxel_vol  # mass in g

    mass = itk.GetImageFromArray(amass)
    mass.CopyInformation(ct_itk)

    return mass


class IsotopeBuilder:
    """
    A description of a G4Isotope that can be build.
    """

    def __init__(self, material_database):
        self.type = "isotope"
        self.name = None
        self.Z = None
        self.N = None
        self.A = None
        self.material_database = material_database

    def __repr__(self):
        u = g4_units.g_mole
        s = f"({self.type}) {self.name} Z={self.Z} N={self.N} A={self.A / u} g/mole"
        return s

    def read(self, line):
        self.type = "isotope"
        s = line.split(":")
        # name
        self.name = s[0]
        s = s[1].strip()
        s = s.split(";")
        # Z
        self.Z = int(read_tag(s[0], "Z"))
        # N
        self.N = int(read_tag(s[1], "N"))
        # A with units
        self.A = read_tag_with_unit(s[2], "A")

    def build(self):
        m = g4.G4Isotope(self.name, self.Z, self.N, self.A)
        return m


class IsotopicElementBuilder:
    """
    A description of an element created from isotopes, that will can be built on demand.
    An element is described by a list of components that can be isotopes or sub-elements.
    """

    def __init__(self, material_database):
        self.type = "isotope"
        self.name = None
        self.symbol = None
        self.n = None
        self.components = {}
        self.material_database = material_database

    def __repr__(self):
        s = f"({self.type}) {self.name} {self.n} {self.components}"
        return s

    def read(self, f, line):
        # read the name
        s = line.split(":")
        if len(s) != 2:
            fatal(
                f"Error line {line}, expecting an element name follow by a colon ':'."
            )
        name = s[0]
        self.name = name

        # reading n and symbol
        s = s[1].split(";")
        if len(s) != 2:
            fatal(f"Error while parsing element {self.name}, line {line}")

        # nb of components
        self.n = int(read_tag(s[0], "n"))

        # symbol
        self.symbol = read_tag(s[1], "S")

        # elements
        for e in range(self.n):
            line = read_next_line(f)
            if line.startswith("+iso"):
                e = self.read_one_isotope(line)
                self.components[e.name] = e

    def read_one_isotope(self, line):
        # skip the initial +iso
        s = line.split("+iso:")
        s = re.split("[;,]", s[1])
        if len(s) != 2:
            fatal(f"Error while reading the line: {line} \n" f'Expected "name=" ; "f="')
        # read the name
        elname = read_tag(s[0], "name")
        if elname == "auto":
            elname = self.name
        if not elname:
            fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        # read f
        f = float(read_tag(s[1], "f"))
        e = Box({"name": elname, "f": f, "type": "isotope"})
        return e

    def build(self):
        m = g4.G4Element(self.name, self.symbol, self.n)
        # add all components
        for iso in self.components.values():
            self.add_iso_to_element(m, iso)
        return m

    def add_iso_to_element(self, elem, iso):
        b = self.material_database.FindOrBuildIsotope(iso.name)
        elem.AddIsotope(b, iso.f)


class ElementBuilder:
    """
    A description of a G4Element that can be build.
    """

    def __init__(self, material_database):
        self.type = "element"
        self.name = None
        self.symbol = None
        self.Zeff = None
        self.Aeff = None
        self.material_database = material_database

    def __repr__(self):
        u = g4_units.g_mole
        s = f"({self.type}) {self.name} ({self.symbol}) Z={self.Zeff} A={self.Aeff / u} g/mole"
        return s

    def read(self, line):
        self.type = "element"
        s = line.split(":")
        # name
        self.name = s[0]
        s = s[1].strip()
        s = s.split(";")
        # symbol
        self.symbol = read_tag(s[0], "S")
        # Z
        self.Zeff = float(read_tag(s[1], "Z"))
        # A with units
        self.Aeff = read_tag_with_unit(s[2], "A")

    def build(self):
        m = g4.G4Element(self.name, self.symbol, self.Zeff, self.Aeff)
        return m


class MaterialBuilder:
    """
    A description of a material, that will can be built on demand.
    A material is described by a list of components that can be elements or sub-materials.
    """

    def __init__(self, material_database):
        self.type = "material"
        self.name = None
        self.symbol = None
        self.density = None
        self.n = None
        self.state = None
        self.components = {}
        self.material_database = material_database

    def __repr__(self):
        s = f"({self.type}) {self.name} {self.density} {self.n} {self.components}"
        return s

    def read(self, f, line):
        # read the name
        s = line.split(":")
        if len(s) != 2:
            fatal(
                f"Error line {line}, expecting a material name follow by a colon ':'."
            )
        name = s[0]
        self.name = name

        # reading density, n, state
        s = s[1].split(";")
        if len(s) != 3 and len(s) != 2:
            fatal(f"Error while parsing material {self.name}, line {line}")

        # density
        self.density = read_tag_with_unit(s[0], "d")
        if not self.density:
            fatal(
                f"Error while parsing material {self.name}, line {line}\n"
                f'Expected density with "d=XXX"'
            )

        # nb of components
        self.n = int(read_tag(s[1], "n"))

        # state
        if len(s) > 2:
            self.state = read_tag(s[2], "state")
            if self.state:
                self.state = self.state.lower()

        # elements
        for e in range(self.n):
            line = read_next_line(f)
            if line.startswith("+mat:"):
                e = self.read_one_submat(line)
                self.components[e.name] = e
            if line.startswith("+el"):
                e = self.read_one_element(line)
                self.components[e.name] = e

    def read_one_element(self, line):
        # skip the initial +el
        s = line.split("+el:")
        s = re.split("[;,]", s[1])
        if len(s) != 2:
            fatal(
                f"Error while reading the line: {line} \n"
                f'Expected "name=" ; "n=" or "f="'
            )
        # read the name
        elname = read_tag(s[0], "name")
        if elname == "auto":
            elname = self.name
        if not elname:
            fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        # read f or n, put the other one to 'None'
        f = None
        n = read_tag(s[1], "n")
        if not n:
            f = float(read_tag(s[1], "f"))
            if f == 0:
                fatal(
                    f"Error during reading material database {self.material_database.current_filename}"
                    f", for the sub material {elname}, the fraction 'f=' is 0."
                )
        else:
            n = int(n)
        e = Box({"name": elname, "n": n, "f": f, "type": "element"})
        return e

    def read_one_submat(self, line):
        # skip the initial +mat
        s = line.split("+mat:")
        s = re.split("[;,]", s[1])
        if len(s) != 2:
            fatal(
                f"Error while reading the line: {line} \n"
                f'Expected "name=" ; "n=" or "f="'
            )
        # read the name
        elname = read_tag(s[0], "name")
        if not elname:
            fatal(
                f"Error reading line {line} \n during the elements of material {self.name}"
            )
        # read f
        f = read_tag(s[1], "f")
        if f is not None:
            f = float(f)
        else:
            fatal(
                f"Error during reading material database {self.material_database.current_filename}"
                f", for the sub material {elname}, except fraction 'f=', while the line is {line}"
            )

        # build the dict
        e = Box({"name": elname, "n": None, "f": f, "type": "material"})
        return e

    def build(self):
        switcher = {
            None: g4.G4State.kStateUndefined,
            "solid": g4.G4State.kStateSolid,
            "liquid": g4.G4State.kStateLiquid,
            "gaz": g4.G4State.kStateGas,
            "gas": g4.G4State.kStateGas,
        }
        state = switcher.get(self.state, f"Invalid material state {self.state}")

        # default temp
        kelvin = g4_units.kelvin
        temp = 293.15 * kelvin

        # default pressure
        atmosphere = g4_units.atmosphere
        pressure = 1 * atmosphere

        # compute the correct nb of elements
        n = 0
        for elem in self.components.values():
            if elem.type == "element":
                n += 1
            else:
                subm = self.material_database.FindOrBuildMaterial(elem.name)
                n += len(subm.GetElementVector())

        # create material
        mat = g4.G4Material(self.name, self.density, n, state, temp, pressure)

        # add all components
        for elem in self.components.values():
            if elem.type == "element":
                self.add_element_to_material(mat, elem)
            else:
                self.add_submat_to_material(mat, elem)
        return mat

    def add_element_to_material(self, mat, elem):
        b = self.material_database.FindOrBuildElement(elem.name)
        if elem.f is None:
            mat.AddElementByNumberOfAtoms(b, elem.n)  # FIXME AddElementByMassFraction
        else:
            mat.AddElementByMassFraction(b, elem.f)

    def add_submat_to_material(self, mat, elem):
        subm = self.material_database.FindOrBuildMaterial(elem.name)
        elems = subm.GetElementVector()
        i = 0
        for el in elems:
            elf = subm.GetElementFraction(i)
            f = elf * elem.f
            name = str(el.GetName())
            b = self.material_database.FindOrBuildElement(name)
            mat.AddElementByMassFraction(b, f)
            i += 1


class MaterialDatabase:
    """
    Manage a unique list of Geant4 materials and elements.
    The materials/elements are read in a DB txt file or in NIST.
    They are only build on demand, during the geometry Construct
    """

    def __init__(self):
        # list of all db filenames where to read material
        self.filenames = []
        # list of all read material (not build)
        self.material_builders = {}
        self.material_builders_by_filename = {}
        # list of all read element (not build)
        self.element_builders = {}
        self.element_builders_by_filename = {}
        # list of all read isotope (not build)
        self.isotope_builders = {}
        self.isotope_builders_by_filename = {}
        # additional manually added materials
        self.new_materials_nb_atoms = {}
        self.new_materials_weights = {}
        # built materials
        self.g4_materials = {}
        # built elements
        self.g4_elements = {}
        # built isotopes
        self.g4_isotopes = {}
        # internal state when reading
        self.current_section = None
        self.current_filename = None
        # specific to NIST materials
        self.g4_NistManager = None
        self.nist_material_names = None
        self.nist_element_names = None

    def __getstate__(self):
        return_dict = self.__dict__
        # remove items that cannot be pickled, e.g. G4 objects
        return_dict["g4_materials"] = {}
        return_dict["g4_elements"] = {}
        return_dict["g4_NistManager"] = None
        return_dict["nist_material_names"] = None
        return_dict["nist_element_names"] = None
        return_dict["material_builders_by_filename"].pop("NIST", None)
        return_dict["element_builders_by_filename"].pop("NIST", None)
        return return_dict

    def __setstate__(self, state):
        self.__dict__ = state
        self.initialize()

    def read_from_file(self, filename):
        self.filenames.append(filename)
        self.current_filename = filename
        self.element_builders_by_filename[self.current_filename] = {}
        self.material_builders_by_filename[self.current_filename] = {}
        self.isotope_builders_by_filename[self.current_filename] = {}
        with open(filename, "r") as f:
            line = f.readline()
            while line:
                line = line.strip().replace("\t", " ")
                self.read_one_item(f, line)
                line = f.readline()

    def read_one_item(self, f, line):
        # skip empty lines
        if len(line) < 1:
            return
        # skip comment line
        if line[0] == "#":
            return
        # check if the current section change
        w = line.split()[0]
        if w == "[Isotopes]":
            self.current_section = "isotope"
            return
        if w == "[Elements]":
            self.current_section = "element"
            return
        if w == "[Materials]":
            self.current_section = "material"
            return
        if not self.current_section:
            fatal(
                f"Error while reading the file {self.current_filename}, "
                f"current section is {self.current_section}. "
                f"File must start with [Isotopes], [Elements] or [Materials]"
            )
        if self.current_section == "isotope":
            b = IsotopeBuilder(self)
            b.read(line)
            self.isotope_builders[b.name] = b
            self.isotope_builders_by_filename[self.current_filename][b.name] = b
        if self.current_section == "element":
            b = None
            if "Z=" in line:
                b = ElementBuilder(self)
                b.read(line)
            if "n=" in line:
                b = IsotopicElementBuilder(self)
                b.read(f, line)
            self.element_builders[b.name] = b
            self.element_builders_by_filename[self.current_filename][b.name] = b
        if self.current_section == "material":
            b = MaterialBuilder(self)
            b.read(f, line)
            self.material_builders[b.name] = b
            self.material_builders_by_filename[self.current_filename][b.name] = b

    def init_NIST(self):
        if self.g4_NistManager is None:
            self.g4_NistManager = g4.G4NistManager.Instance()
            self.nist_material_names = self.g4_NistManager.GetNistMaterialNames()
            self.nist_element_names = self.g4_NistManager.GetNistElementNames()
            self.material_builders_by_filename["NIST"] = self.nist_material_names
            self.element_builders_by_filename["NIST"] = self.nist_element_names

    # FIXME: make arguments explicit
    def add_material_nb_atoms(self, *args):
        """
        Usage example:
        "Lead", ["Pb"], [1], 11.4 * gcm3
        "BGO", ["Bi", "Ge", "O"], [4, 3, 12], 7.13 * gcm3)
        """
        name = args[0]
        self.new_materials_nb_atoms[name] = args

    # FIXME: make arguments explicit
    def add_material_weights(self, *args):
        """
        Usage example :
        add_material_weights(name, elems_symbol_nz, weights_nz, 3 * gcm3)
        """
        name = args[0]
        self.new_materials_weights[name] = args

    def initialize(self):
        self.init_NIST()
        self.init_user_mat()

    def init_user_mat(self):
        for mat_name in self.new_materials_nb_atoms:
            if mat_name in self.g4_materials:
                fatal(f"Material {mat_name} is already constructed")
            mat_info = self.new_materials_nb_atoms[mat_name]
            try:
                mat = self.g4_NistManager.ConstructNewMaterialNbAtoms(*mat_info)
            except:  # FIXME: this should specify the exception to catch
                fatal(f"Cannot construct the material (nb atoms): {mat_info}")
            self.g4_materials[mat_name] = mat
        self.new_materials_nb_atoms = []
        for mat_name in self.new_materials_weights:
            if mat_name in self.g4_materials:
                fatal(f"Material {mat_name} is already constructed")
            mat_info = self.new_materials_weights[mat_name]
            try:
                mat = self.g4_NistManager.ConstructNewMaterialWeights(*mat_info)
            except:
                fatal(f"Cannot construct the material (weights): {mat_info}")
            self.g4_materials[mat_name] = mat
        self.new_materials_weights = []

    def FindOrBuildIsotope(self, isotope_name):
        # try to build the isotope if it does not yet exist
        if isotope_name not in self.g4_materials:
            if isotope_name in self.isotope_builders:
                self.g4_isotopes[isotope_name] = self.isotope_builders[
                    isotope_name
                ].build()
            else:
                fatal(f'Cannot find nor build isotope named "{isotope_name}"')
        return self.g4_isotopes[isotope_name]

    def FindOrBuildMaterial(self, material_name):
        self.init_NIST()
        self.init_user_mat()
        # try to build the material if it does not yet exist
        if material_name not in self.g4_materials:
            if material_name in self.nist_material_names:
                self.g4_materials[material_name] = (
                    self.g4_NistManager.FindOrBuildMaterial(material_name)
                )
            elif material_name in self.material_builders:
                self.g4_materials[material_name] = self.material_builders[
                    material_name
                ].build()
            else:
                fatal(f'Cannot find nor build material named "{material_name}"')
        return self.g4_materials[material_name]

    def FindOrBuildElement(self, element_name):
        self.init_NIST()
        # return if already exist
        if element_name in self.g4_elements:
            return self.g4_elements[element_name]
        # we build and store the G4 element if not
        if element_name in self.nist_element_names:
            be = self.g4_NistManager.FindOrBuildElement(element_name)
            self.g4_elements[element_name] = be
            return be
        if element_name not in self.element_builders:
            fatal(f'Cannot find nor build element named "{element_name}"')
        be = self.element_builders[element_name].build()
        self.g4_elements[element_name] = be
        return be

    def get_database_material_names(self, db=None):
        if db is None:
            names = [m for m in self.material_builders]
            return names
        if db not in self.material_builders_by_filename:
            fatal(
                f"The database '{db}' is not in the list of read database: {self.filenames}"
            )
        return [name for name in self.material_builders_by_filename[db]]


def write_material_database(sim, materials, filename):
    fn = str(filename)
    with open(fn, "w") as file:
        file.write("[Materials]\n")
        for mat in materials:
            m = sim.volume_manager.material_database.FindOrBuildMaterial(mat)
            s = dump_material_like_Gate(m)
            file.write(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test037_pet_hits_singles_helpers as t37
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test037_pet", "test037")

    """
    This test considers a PET system (Vereos Philips), with NEMA NECR linear fantom and source (F18).
    The digitizer is simplified to:
        1) hits collection
        2) singles are obtained with one simple adder (EnergyWeightedCentroidPosition)

    Note that this is not a correct digitizer (no blurring, no noise, no dead-time, etc).

    Hits are recorded into the crystal volumes (repeated 23,040 times).
    Singles are created by grouping hits from the same event, in the same crystal.

    The output is a root file composed of two trees 'Hits' and 'Singles'.
    Both are compared to an equivalent legacy Gate simulation.

    Salvadori J, Labour J, Odille F, Marie PY, Badel JN, Imbert L, Sarrut D.
    Monte Carlo simulation of digital photon counting PET.
    EJNMMI Phys. 2020 Apr 25;7(1):23.
    doi: 10.1186/s40658-020-00288-w

    """

    # create the simulation
    sim = gate.Simulation()
    crystal = t37.create_pet_simulation(sim, paths)
    t37.add_digitizer(sim, paths, "1", crystal)

    # timing
    sec = gate.g4_units.second
    sim.run_timing_intervals = [[0, 0.00005 * sec]]

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # ----------------------------------------------------------------------------------------------------------

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    p = paths.gate / "output"
    stats_ref = utility.read_stat_file(p / "stats1.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.028)

    # check root hits
    hc = sim.get_actor("Hits")
    f = p / "output1.root"
    is_ok = t37.check_root_hits(paths, 1, f, hc.get_output_path()) and is_ok

    # check root singles
    sc = sim.get_actor("Singles")
    is_ok = t37.check_root_singles(paths, 1, f, sc.get_output_path()) and is_ok

    utility.test_ok(is_ok)



# if pytomography is not installed, we ignore this module
# this is needed for test080_check_classes_are_processed.py
# that check all modules
try:
    import pytomography
except ModuleNotFoundError:
    print("pytomography module is not installed. Skipping pytomography_helpers.")
    import sys

    # Unload this module to prevent errors in other imports
    sys.modules[__name__] = None
    raise SystemExit

from opengate.exception import warning
from opengate.contrib.spect.spect_config import *
from opengate.image import resample_itk_image
from opengate.contrib.dose.photon_attenuation_image_helpers import (
    create_photon_attenuation_image,
)

from pytomography.metadata.SPECT import SPECTObjectMeta, SPECTProjMeta, SPECTPSFMeta
from pytomography.transforms.SPECT import SPECTPSFTransform, SPECTAttenuationTransform
from pytomography.projectors.SPECT import SPECTSystemMatrix
from pytomography.likelihoods import PoissonLogLikelihood
from pytomography.algorithms import OSEM
from pytomography.utils import compute_EW_scatter

import torch
import SimpleITK as sitk
import numpy as np
import json
import os
from pathlib import Path


def rotate_np_image_to_pytomography(np_image):
    """
    Rotate a numpy image to the pytomography coordinate system.
    For example, for the attenuation map.
    """
    rotation_arr = np.transpose(np_image, axes=(2, 1, 0))
    return rotation_arr


def rotate_np_pytomography_to_image(np_image):
    """
    Rotate a numpy pytomography reconstructed image to the Gate coordinate system.
    """
    rotation_arr = np.transpose(np_image, axes=(2, 1, 0))
    rotation_arr = rotation_arr[:, ::-1, :].copy()
    return rotation_arr


def rotate_np_sinogram_to_pytomography(np_sinogram):
    """
    Rotate numpy sinogram to the Pytomography coordinate system [angles,z,x] -> [angles,x,z]
    we reverse the Z axis (gantry rotation) because the rotation of the
    gantry is in the opposite direction
    """
    rotated_sinogram = np.transpose(np_sinogram, axes=(0, 2, 1))
    rotated_sinogram = rotated_sinogram[:, :, ::-1].copy()
    return rotated_sinogram


def pytomography_read_metadata(json_file):
    json_file = Path(json_file).resolve()
    with open(json_file, "r") as f:
        metadata = json.load(f)
        metadata["folder"] = json_file.parent
    return metadata


def pytomography_new_metadata():
    return {
        "Energy Window Data": {
            "energy_window_lower_bounds": [],
            "energy_window_upper_bounds": [],
            "number_of_energy_windows": 0,
        },
        "Detector Physics Data": {
            # some default values as an example
            "hole_shape": 6,
            "hole_diameter": 0.294,
            "hole_spacing": 0.4,
            "collimator_thickness": 4.064,
            "collimator_material": "lead",
            "crystal_spatial_resolution_fwhm_energy_model": "A*sqrt(B/energy)",
            "crystal_spatial_resolution_fwhm_energy_model_parameters": [0.36, 140.5],
            "crystal_energy_resolution_fwhm_pct_energy_model": "A*sqrt(B/energy)",
            "crystal_energy_resolution_fwhm_pct_energy_model_parameters": [9.9, 140.5],
            "crystal_width": 55.3,
            "crystal_height": 38.1,
            "isotope_name": "Lu177",
        },
        "Attenuation Data": {
            "source_file": None,
            "voxel_size_x": 0.24,
            "voxel_size_y": 0.24,
            "voxel_size_z": 0.24,
            "dimension_x": 256,
            "dimension_y": 256,
            "dimension_z": 256,
            "origin_x": -30.599999999999998,
            "origin_y": -30.599999999999998,
            "origin_z": -30.599999999999998,
            "energy": 208,
        },
        "Projection Geometry Data": {
            "source_file": None,
            "number_of_projections": 0,
            "detector_position_x": [],
            "detector_position_y": [],
            "detector_position_z": [],
            "detector_orientation_x": [],
            "detector_orientation_y": [],
            "detector_orientation_z": [],
            "projection_pixel_size_r": 0.48,
            "projection_pixel_size_z": 0.48,
            "projection_dimension_r": 128,
            "projection_dimension_z": 128,
            "starting_angle_deg": -90,
        },
    }


def pytomography_add_detector_orientation(metadata, detector):
    geom = metadata["Projection Geometry Data"]
    for dp in detector.dynamic_params.values():
        translations = np.array(dp["translation"]) / g4_units.cm
        rotations = dp["rotation"]
        for tr, rot in zip(translations, rotations):
            # translation and rotation are according to the mother volume of the detector
            geom["detector_position_x"].append(tr[0])
            geom["detector_position_y"].append(tr[1])
            # detector_orientation_z is *not* used yet by pytomography
            geom["detector_position_z"].append(tr[2])
            # detector_orientation is *not* used yet by pytomography
            geom["detector_orientation_x"].append(rot[0].tolist())
            geom["detector_orientation_y"].append(rot[1].tolist())
            geom["detector_orientation_z"].append(rot[2].tolist())
        geom["number_of_projections"] += len(translations)


def pytomography_set_detector_info(metadata, pixel_spacing, dimension):
    geom = metadata["Projection Geometry Data"]
    geom["projection_pixel_size_r"] = pixel_spacing[0] / g4_units.cm
    geom["projection_pixel_size_z"] = pixel_spacing[1] / g4_units.cm
    geom["projection_dimension_r"] = dimension[0]
    geom["projection_dimension_z"] = dimension[1]


def pytomography_set_energy_windows(metadata, channels):
    ew = metadata["Energy Window Data"]
    for channel in channels:
        ew["energy_window_lower_bounds"].append(channel["min"])
        ew["energy_window_upper_bounds"].append(channel["max"])
        ew["number_of_energy_windows"] += 1


def pytomography_create_sinogram(filenames, number_of_angles, output_filename):
    # consider sinogram for all energy windows
    sinograms = read_projections_as_sinograms(filenames, number_of_angles)
    sino_arr = None
    for sinogram in sinograms:
        arr = sitk.GetArrayFromImage(sinogram)
        arr = rotate_np_sinogram_to_pytomography(arr)
        # concatenate all energy windows
        if sino_arr is None:
            sino_arr = arr
        else:
            sino_arr = np.concatenate((sino_arr, arr), axis=0)

    sino_itk = sitk.GetImageFromArray(sino_arr)
    sino_itk.SetSpacing(sinograms[0].GetSpacing())
    sino_itk.SetOrigin(sinograms[0].GetOrigin())
    sino_itk.SetDirection(sinograms[0].GetDirection())
    sitk.WriteImage(sino_itk, output_filename)
    return sino_arr


def create_attenuation_image(
    ct_filename,
    energy,
    attenuation_filename,
    size,
    spacing,
    translation=None,
    verbose=False,
):
    # resample
    image = itk.imread(ct_filename)
    verbose and print(f"Resample CT image {ct_filename} to {size} and {spacing}")
    image = resample_itk_image(
        image,
        size,
        spacing,
        default_pixel_value=-1000,
        linear=False,
        translation=translation,
    )
    itk.imwrite(image, attenuation_filename)

    # mumap
    verbose and print(f"Compute attenuation map for E={energy/g4_units.MeV} MeV")
    attenuation_image = create_photon_attenuation_image(
        attenuation_filename,
        labels_filename=None,
        energy=energy,
        material_database=None,
        database="EPDL",
        verbose=False,
        density_tol=0.05 * gate.g4_units.g_cm3,
    )
    itk.imwrite(attenuation_image, attenuation_filename)


def pytomography_set_attenuation_data(
    metadata,
    energy,
    attenuation_filename,
    translation=None,
    verbose=False,
):

    # check attenuation_filename must be mhd/raw
    attenuation_filename = Path(attenuation_filename)
    if attenuation_filename.suffix != ".mhd":
        fatal(
            f"Attenuation file must be mhd/raw, while the "
            f"extension is {attenuation_filename.suffix} ({attenuation_filename})"
        )
    raw_filename = attenuation_filename.with_suffix(".raw")

    # read attenuation image
    attenuation_image = sitk.ReadImage(attenuation_filename)
    size = attenuation_image.GetSize()
    spacing = attenuation_image.GetSpacing()

    # information and translation
    if translation is None:
        translation = np.array([0, 0, 0])
    else:
        translation = np.array(translation)
    verbose and print(f"Phantom translation: {translation}")
    ad = metadata["Attenuation Data"]
    ad["source_file"] = str(raw_filename.name)
    ad["energy"] = energy / g4_units.keV  # this is not used by pytomography
    ad["voxel_size_x"] = spacing[0] / g4_units.cm
    ad["voxel_size_y"] = spacing[1] / g4_units.cm
    ad["voxel_size_z"] = spacing[2] / g4_units.cm
    ad["dimension_x"] = size[0]
    ad["dimension_y"] = size[1]
    ad["dimension_z"] = size[2]
    ad["origin_x"] = attenuation_image.GetOrigin()[0] / g4_units.cm
    ad["origin_y"] = attenuation_image.GetOrigin()[1] / g4_units.cm
    ad["origin_z"] = attenuation_image.GetOrigin()[2] / g4_units.cm
    ad["translation"] = (translation / g4_units.cm).tolist()


# ------ below should be in pytomography ? --------


def pytomography_get_detector_data(metadata):
    geometry_meta = metadata["Projection Geometry Data"]
    detector_position_x = np.array(geometry_meta["detector_position_x"])
    detector_position_y = np.array(geometry_meta["detector_position_y"])

    # the following are not used yet
    detector_position_z = np.array(geometry_meta["detector_position_z"])
    detector_orientation_x = np.array(geometry_meta["detector_orientation_x"])
    detector_orientation_y = np.array(geometry_meta["detector_orientation_y"])
    detector_orientation_z = np.array(geometry_meta["detector_orientation_z"])

    # pixel size and spacing
    projection_pixel_size_r = geometry_meta["projection_pixel_size_r"]
    projection_pixel_size_z = geometry_meta["projection_pixel_size_z"]
    projection_dimension_r = geometry_meta["projection_dimension_r"]
    projection_dimension_z = geometry_meta["projection_dimension_z"]

    # For now, assume all oriented towards the centre
    radii = np.sqrt(detector_position_x**2 + detector_position_y**2)
    angles = np.arctan2(detector_position_y, detector_position_x)

    # starting angle
    angles = np.degrees(angles) + geometry_meta["starting_angle_deg"]

    object_meta = SPECTObjectMeta(
        [projection_pixel_size_r, projection_pixel_size_r, projection_pixel_size_z],
        [projection_dimension_r, projection_dimension_r, projection_dimension_z],
    )

    proj_meta = SPECTProjMeta(
        [projection_dimension_r, projection_dimension_z],
        [projection_pixel_size_r, projection_pixel_size_z],
        angles,
        radii,
    )

    return object_meta, proj_meta


def pytomography_read_projections(metadata, folder=None):
    object_meta, proj_meta = pytomography_get_detector_data(metadata)
    energy_meta = metadata["Energy Window Data"]
    number_of_energy_windows = energy_meta["number_of_energy_windows"]
    geometry_meta = metadata["Projection Geometry Data"]
    image_file = geometry_meta["source_file"]
    if folder is None:
        folder = Path(metadata["folder"])
    projections = np.fromfile(folder / image_file, dtype=np.float32)
    projections = projections.reshape(number_of_energy_windows, *proj_meta.shape)
    projections = torch.tensor(projections).to(pytomography.device)
    return projections


def get_psf_meta_from_json(metadata, photon_energy, min_sigmas=3):
    FWHM2sigma = 1 / (2 * np.sqrt(2 * np.log(2)))
    detector_meta = metadata["Detector Physics Data"]
    hole_diameter = detector_meta["hole_diameter"]
    hole_length = detector_meta["collimator_thickness"]
    crystal_spatial_resolution_model = detector_meta[
        "crystal_spatial_resolution_fwhm_energy_model"
    ]
    if crystal_spatial_resolution_model == "A*sqrt(B/energy)":
        crystal_spatial_resolution_model_parameters = detector_meta[
            "crystal_spatial_resolution_fwhm_energy_model_parameters"
        ]
        A = crystal_spatial_resolution_model_parameters[0]
        B = crystal_spatial_resolution_model_parameters[1]
        intrinsic_resolution = A * np.sqrt(B / photon_energy)
    else:
        raise ValueError("Unsupported crystal spatial resolution model")
    collimator_slope = hole_diameter / hole_length * FWHM2sigma
    collimator_intercept = hole_diameter * FWHM2sigma
    sigma_fit = lambda r, a, b, c: np.sqrt((a * r + b) ** 2 + c**2)
    sigma_fit_params = [collimator_slope, collimator_intercept, intrinsic_resolution]
    return SPECTPSFMeta(
        sigma_fit_params=sigma_fit_params, sigma_fit=sigma_fit, min_sigmas=min_sigmas
    )


def compute_tew_scatter_estimate(
    metadata,
    index_lower,
    index_upper,
    index_peak,
    weighting_lower: float = 0.5,
    weighting_upper: float = 0.5,
    sigma_theta: float = 0,
    sigma_r: float = 0,
    sigma_z: float = 0,
    N_sigmas: int = 3,
    return_scatter_variance_estimate: bool = False,
):
    energy_data = metadata["Energy Window Data"]
    energy_window_lower_bounds = energy_data["energy_window_lower_bounds"]
    energy_window_upper_bounds = energy_data["energy_window_upper_bounds"]
    width_lower = (
        energy_window_upper_bounds[index_lower]
        - energy_window_lower_bounds[index_lower]
    )
    width_upper = (
        energy_window_upper_bounds[index_upper]
        - energy_window_lower_bounds[index_upper]
    )
    width_peak = (
        energy_window_upper_bounds[index_peak] - energy_window_lower_bounds[index_peak]
    )
    projections = pytomography_read_projections(metadata)
    projections_lower = projections[index_lower]
    projections_upper = projections[index_upper]
    _, proj_meta = pytomography_get_detector_data(metadata)
    TEW = compute_EW_scatter(
        projections_lower,
        projections_upper,
        width_lower,
        width_upper,
        width_peak,
        weighting_lower,
        weighting_upper,
        proj_meta=proj_meta,
        sigma_r=sigma_r,
        sigma_z=sigma_z,
        sigma_theta=sigma_theta,
        N_sigmas=N_sigmas,
        return_scatter_variance_estimate=return_scatter_variance_estimate,
    )
    return TEW


def compute_dew_scatter_estimate(
    metadata,
    index_scatter_window,  # Index of the single scatter window (e.g., lower scatter)
    index_peak_window,  # Index of the photopeak window
    k_factor: float = 0.5,  # The crucial 'k' factor for DEW. Default is illustrative.
    sigma_theta: float = 0,
    sigma_r: float = 0,
    sigma_z: float = 0,
    n_sigmas: int = 3,
    return_scatter_variance_estimate: bool = False,  # DEW typically doesn't yield variance in this way
):
    """
    Computes the scatter estimate in the photopeak window using the Dual Energy Window (DEW) method.

    This function implements the classic DEW approach where scatter in the photopeak is estimated
    by scaling the counts in a single scatter window by a calibrated 'k' factor.
    It's particularly suited for radionuclides like Tc-99m, though the k_factor is crucial.

    Args:
        metadata (dict): A dictionary containing metadata, including 'Energy Window Data'
                         with 'energy_window_lower_bounds' and 'energy_window_upper_bounds'.
                         Also implicitly contains information for reading projections.
        index_scatter_window (int): The index of the single scatter window in the metadata's
                                    energy window data. This is typically a lower scatter window.
        index_peak_window (int): The index of the photopeak window in the metadata's
                                 energy window data.
        k_factor (float): The empirically determined or simulated 'k' factor used to scale
                          the scatter window counts to estimate photopeak scatter.
                          This value is critical and depends on the acquisition setup.
                          A common range for Tc-99m might be 0.4 to 0.6, but needs calibration.
        sigma_theta (float): Standard deviation for angular Gaussian smoothing (in radians).
                             Applied to the scatter window projections before scaling.
        sigma_r (float): Standard deviation for radial Gaussian smoothing (in pixels).
                         Applied to the scatter window projections before scaling.
        sigma_z (float): Standard deviation for axial Gaussian smoothing (in pixels).
                         Applied to the scatter window projections before scaling.
        n_sigmas (int): Number of standard deviations for Gaussian smoothing kernel truncation.
        return_scatter_variance_estimate (bool): Not typically used for direct DEW,
                                                 included for function signature consistency.
                                                 Will always return None if True.

    Returns:
        numpy.ndarray: A 3D NumPy array representing the estimated scatter in the photopeak window,
                       with the same shape as the photopeak projections.
    """

    energy_data = metadata["Energy Window Data"]
    energy_window_lower_bounds = energy_data["energy_window_lower_bounds"]
    energy_window_upper_bounds = energy_data["energy_window_upper_bounds"]

    # Although not directly used for the DEW calculation formula,
    # it's good practice to ensure these indices are valid.
    # We'll use the 'width_peak' for potential future extensions or consistency.
    width_scatter = (
        energy_window_upper_bounds[index_scatter_window]
        - energy_window_lower_bounds[index_scatter_window]
    )
    width_peak = (
        energy_window_upper_bounds[index_peak_window]
        - energy_window_lower_bounds[index_peak_window]
    )

    # Read all projections from the metadata
    projections = pytomography_read_projections(metadata)

    # Get the specific scatter window projections
    projections_scatter = projections[index_scatter_window]

    # Get detector metadata for smoothing
    _, proj_meta = pytomography_get_detector_data(metadata)

    # Apply Gaussian smoothing to the scatter window projections if sigmas are provided
    if sigma_theta > 0 or sigma_r > 0 or sigma_z > 0:
        # Assuming pytomography.utils.gaussian_filter_projections is available
        # or you'd apply it directly using scipy.ndimage.gaussian_filter
        # As there's no direct equivalent to proj_meta handling in scipy.ndimage
        # for these specific (theta, r, z) dimensions, we'll need to define
        # how `pytomography_gaussian_filter_projections` works.
        # For simplicity, assuming a conceptual helper or direct `scipy.ndimage` application.

        # If pytomography provides a direct way to smooth based on sigma_r, sigma_z, sigma_theta,
        # you'd use that. Otherwise, you might apply standard N-D Gaussian filtering,
        # being mindful of the axes.
        # Example using a conceptual pytomography smoothing function:
        # projections_scatter_smoothed = pytomography_gaussian_filter_projections(
        #     projections_scatter,
        #     sigma_r=sigma_r,
        #     sigma_z=sigma_z,
        #     sigma_theta=sigma_theta,
        #     N_sigmas=N_sigmas,
        #     proj_meta=proj_meta # Pass proj_meta for proper dimension handling
        # )
        # For a practical example with standard scipy, you'd map sigma_r, sigma_z, sigma_theta
        # to the correct axes of the projections_scatter array.
        # Projections are typically (angle, radial_bin, axial_bin).
        # Let's use `pytomography.utils.gaussian_filter_projections` if it's what's expected.
        try:
            from pytomography.utils import gaussian_filter_projections

            projections_scatter_smoothed = gaussian_filter_projections(
                projections_scatter,
                sigma_r=sigma_r,
                sigma_z=sigma_z,
                sigma_theta=sigma_theta,
                N_sigmas=n_sigmas,
                proj_meta=proj_meta,  # Pass proj_meta for proper dimension handling
            )
        except AttributeError:
            print(
                "Warning: `pytomography.utils.gaussian_filter_projections` not found. Skipping smoothing."
            )
            print(
                "Please ensure this function is available or handle smoothing manually (e.g., using scipy.ndimage)."
            )
            projections_scatter_smoothed = projections_scatter
    else:
        projections_scatter_smoothed = projections_scatter

    # Implement the DEW scatter estimation: Scatter_peak = k_factor * Scatter_window
    DEW_scatter_estimate = k_factor * projections_scatter_smoothed

    # DEW typically doesn't provide a variance estimate directly in this way.
    # So we return None for consistency with the TEW function signature if requested.
    if return_scatter_variance_estimate:
        return DEW_scatter_estimate, None
    else:
        return DEW_scatter_estimate


def get_attenuation_map_from_json(metadata):
    attenuation_meta = metadata["Attenuation Data"]
    dimension_x = attenuation_meta["dimension_x"]
    dimension_y = attenuation_meta["dimension_y"]
    dimension_z = attenuation_meta["dimension_z"]

    # the following are not used for the moment
    voxel_size_x = attenuation_meta["voxel_size_x"]
    voxel_size_y = attenuation_meta["voxel_size_y"]
    voxel_size_z = attenuation_meta["voxel_size_z"]
    origin_x = attenuation_meta["origin_x"]
    origin_y = attenuation_meta["origin_y"]
    origin_z = attenuation_meta["origin_z"]

    # the energy is not used
    energy = attenuation_meta["energy"]

    # get the image from the raw data
    imagefile = attenuation_meta["source_file"]
    if imagefile is None:
        return None
    amap = np.fromfile(os.path.join(metadata["folder"], imagefile), dtype=np.float32)
    amap = amap.reshape((dimension_x, dimension_y, dimension_z))

    # rotation from gate to pytomo
    amap = rotate_np_image_to_pytomography(amap)

    # we consider the attenuation map has already been resampled
    # like the projection / reconstructed object
    return torch.tensor(amap).to(pytomography.device)


def pytomography_osem_reconstruction(
    metadata,
    index_peak,
    psf_photon_energy,
    attenuation_flag,
    scatter_mode=None,
    index_lower=None,
    index_upper=None,
    n_iters=4,
    n_subsets=8,
):
    # get metadata information
    object_meta, proj_meta = pytomography_get_detector_data(metadata)

    # get projections data
    projections = pytomography_read_projections(metadata)

    # attenuation correction
    att_transform = None
    if attenuation_flag:
        amap = get_attenuation_map_from_json(metadata)
        if amap is None:
            fatal(f"No attenuation map found in metadata. Aborting.")
        att_transform = SPECTAttenuationTransform(amap)

    # PSF correction
    psf_meta = get_psf_meta_from_json(
        metadata,
        photon_energy=psf_photon_energy / g4_units.keV,
    )
    psf_transform = SPECTPSFTransform(psf_meta)

    # scatter correction (it may be None if no sc correction)
    additive_term = None
    if index_lower is None:
        index_lower = index_peak - 1
    if scatter_mode == "TEW":
        if index_upper is None:
            index_upper = index_peak + 1
        additive_term = compute_tew_scatter_estimate(
            metadata,
            index_lower=index_lower,
            index_upper=index_upper,
            index_peak=index_peak,
        )
    if scatter_mode == "DEW":
        additive_term = compute_dew_scatter_estimate(
            metadata,
            index_scatter_window=index_lower,
            index_peak_window=index_peak,
            k_factor=1,
            sigma_theta=0,
            sigma_r=0,
            sigma_z=0,
            n_sigmas=3,
        )

    # system matrix
    if attenuation_flag:
        obj = [att_transform, psf_transform]
    else:
        obj = [psf_transform]
    system_matrix = SPECTSystemMatrix(
        obj2obj_transforms=obj,
        proj2proj_transforms=[],
        object_meta=object_meta,
        proj_meta=proj_meta,
    )

    # loss function
    likelihood = PoissonLogLikelihood(
        system_matrix=system_matrix,
        projections=projections[index_peak],
        additive_term=additive_term,
    )

    # go !
    recon_algorithm = OSEM(likelihood)
    reconstructed_image = recon_algorithm(n_iters=n_iters, n_subsets=n_subsets)

    # build the final image
    # (warning spacing is in cm)
    spacing = np.array([object_meta.dx, object_meta.dy, object_meta.dz]) * g4_units.cm
    size = np.array(object_meta.shape)
    recon_arr = reconstructed_image.cpu().numpy()
    recon_arr = rotate_np_pytomography_to_image(recon_arr)
    recon_sitk = sitk.GetImageFromArray(recon_arr)
    recon_sitk.SetSpacing(spacing)

    # get the origin according to the attenuation map
    # (warning itk in mm and pytomography in cm)
    am = metadata["Attenuation Data"]
    origin = np.array([am["origin_x"], am["origin_y"], am["origin_z"]])
    origin = origin * g4_units.cm

    recon_sitk.SetOrigin(origin)

    return recon_sitk


def pytomography_build_metadata_and_attenuation_map(
    sc,
    sim,
    attenuation_energy,
    output_folder=None,
    verbose=True,
):
    """
    Here are the necessary inputs to create the metadata
    - list of detector heads: for orientation
    - Digitizer projection actor: for size and spacing
    - Digitizer energy windows actor: for energy window
    - list of projection files: for build the sinogram in pytomography orientation
    - energy for the attenuation map
    """

    # output folder for sinogram and attenuation map
    if output_folder is None:
        output_folder = sc.output_folder

    # create a new (empty) pytomography metadata file
    metadata = pytomography_new_metadata()

    # set the detector orientations
    detectors = sc.detector_config.get_heads(sim)
    for detector in detectors:
        pytomography_add_detector_orientation(metadata, detector)
    nb_proj = metadata["Projection Geometry Data"]["number_of_projections"]
    verbose and print(f"Found a total of {nb_proj} projections")

    # the angle zero is on the side of the table
    metadata["Projection Geometry Data"]["starting_angle_deg"] = -90

    # set the detector size and spacing
    size = [
        sc.detector_config.size[0],
        sc.detector_config.size[1],
        sc.detector_config.size[1],
    ]
    spacing = [
        sc.detector_config.spacing[0],
        sc.detector_config.spacing[1],
        sc.detector_config.spacing[1],
    ]
    pytomography_set_detector_info(metadata, spacing, size)
    verbose and print(f"Found detector size and spacing set to {size} and {spacing}")

    # set the energy windows
    ews = sim.actor_manager.find_actors("energy_window")
    channels = ews[0].channels
    pytomography_set_energy_windows(metadata, channels)
    verbose and print(f"Found {len(channels)} energy windows")

    # create the sinogram with all the projections
    filenames = sc.detector_config.get_proj_filenames(sim)
    o = output_folder / "sinogram.mhd"
    sino = pytomography_create_sinogram(
        filenames, sc.acquisition_config.number_of_angles, o
    )
    verbose and print(f"Build sinogram with shape {sino.shape} in {o}")
    # set the raw data only for pytomography
    metadata["Projection Geometry Data"]["source_file"] = "sinogram.raw"

    # PSF correction "Detector Physics Data"
    dpd = metadata["Detector Physics Data"]
    m = sc.detector_config.get_model_module()
    dpd.update(m.get_pytomography_detector_physics_data(sc.detector_config.collimator))
    dpd["isotope_name"] = sc.source_config.radionuclide

    # crystal mode
    known_crystal_models = ["intevo"]  # , "nm670"]
    if sc.detector_config.model not in known_crystal_models:
        warning(f"Unknown crystal model: {sc.detector_config.model}")
        warning(f'the crystal parameters are set by default like "intevo"')

    if sc.detector_config.model == "intevo":
        dpd["crystal_spatial_resolution_fwhm_energy_model"] = "A*sqrt(B/energy)"
        dpd["crystal_spatial_resolution_fwhm_energy_model_parameters"] = [0.36, 140.5]
        dpd["crystal_energy_resolution_fwhm_pct_energy_model"] = "A*sqrt(B/energy)"
        dpd["crystal_energy_resolution_fwhm_pct_energy_model_parameters"] = [9.9, 140.5]

    if sc.detector_config.model == "nm670":
        dpd["crystal_spatial_resolution_fwhm_energy_model"] = "A*sqrt(B/energy)"
        dpd["crystal_spatial_resolution_fwhm_energy_model_parameters"] = [0.3, 140.5]
        dpd["crystal_energy_resolution_fwhm_pct_energy_model"] = "A*sqrt(B/energy)"
        dpd["crystal_energy_resolution_fwhm_pct_energy_model_parameters"] = [6.3, 140.5]

    # attenuation correction
    if sc.phantom_config.image is not None:
        # read CT image and resample like the reconstructed image
        # for the moment MUST be same size and spacing than the projection images
        # Important: take into account the translation of the phantom!
        create_attenuation_image(
            sc.phantom_config.image,
            attenuation_energy,
            output_folder / "mumap.mhd",
            size,
            spacing,
            translation=sc.phantom_config.translation,
            verbose=True,
        )
        pytomography_set_attenuation_data(
            metadata,
            attenuation_energy,
            output_folder / "mumap.mhd",
            translation=sc.phantom_config.translation,
            verbose=True,
        )
        verbose and print(
            f"Build attenuation map with shape {size} in {output_folder / 'mumap.mhd'}"
        )

    return metadata



import os
import sys

# Some Python versions distributed by Conda have a buggy `os.add_dll_directory`
# which prevents binary wheels from finding the FFmpeg DLLs in the `av.libs`
# directory. We work around this by adding `av.libs` to the PATH.
if os.name == "nt":
    os.environ["PATH"] = (
        os.path.abspath(
            os.path.join(os.path.dirname(__file__), os.pardir, "opengate_core.libs")
        )
        + os.pathsep
        + os.environ["PATH"]
    )
    os.add_dll_directory(
        os.path.join(os.path.dirname(__file__), os.pardir, "opengate_core.libs")
    )
pathCurrentFile = os.path.abspath(__file__)
if sys.platform == "win32":
    print(os.path.dirname(pathCurrentFile))
    os.add_dll_directory(os.path.dirname(pathCurrentFile))

from .opengate_core import *
from .g4DataSetup import *
from .qt5Setup import *

if os.environ.get("GATEONRTD", None) is None:
    check_g4_data()
    set_g4_data_path()
    set_qt5_path()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test042_gauss_gps", "test042"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 10 * cm, 10 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # daughter
    phantom_y = sim.add_volume("Box", "phantom_y")
    phantom_y.mother = phantom.name
    phantom_y.size = [2 * mm, 10 * cm, 2 * mm]
    phantom_y.translation = [49 * mm, 0, 0]
    phantom_y.material = "G4_WATER"
    phantom_y.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_INCLXX_EMZ"
    sim.physics_manager.global_production_cuts.all = 1000 * km
    # FIXME need SetMaxStepSizeInRegion ActivateStepLimiter
    # e.g., like so:
    # sim.physics_manager.set_max_step_size(
    #     volume_name="phantom", max_step_size=1 * mm
    # )
    # or:
    # reg = sim.physics_manager.add_region('reg')
    # reg.max_step_size = 1 * mm
    # reg.associate_volume(phantom)

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 40 * MeV
    source.particle = "proton"
    source.position.type = "disc"  # pos = Beam, shape = circle + sigma
    # rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    # source.position.radius = 8 * mm
    source.position.sigma_x = 8 * mm
    source.position.sigma_y = 8 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.activity = 100 * kBq

    # add dose actor
    dose = sim.add_actor("DoseActor", "doseInXZ")
    dose.output_filename = paths.output / "test042-lateral_xz.mhd"
    dose.attached_to = phantom.name
    dose.size = [250, 1, 250]
    dose.spacing = [0.4, 100, 0.4]
    dose.hit_type = "random"

    dose = sim.add_actor("DoseActor", "doseInXY")
    dose.output_filename = paths.output / "test042-lateral_xy.mhd"
    dose.attached_to = phantom.name
    dose.size = [250, 250, 1]
    dose.spacing = [0.4, 0.4, 100]
    dose.hit_type = "random"

    dose = sim.add_actor("DoseActor", "doseInYZ")
    dose.output_filename = paths.output / "test042-lateral_yz.mhd"
    dose.attached_to = phantom.name
    dose.size = [1, 250, 250]
    dose.spacing = [100, 0.4, 0.4]
    dose.hit_type = "random"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    dose = sim.get_actor("doseInXZ")
    print(dose)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    print()
    gate.exception.warning("Tests stats file")
    stats_ref = utility.read_stat_file(paths.gate_output / "stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.14)

    print()
    gate.exception.warning("Difference for EDEP XZ")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "lateral_xz_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
            paths.output / sim.get_actor("doseInXZ").get_output_path("edep"),
            stats,
            tolerance=10,
            ignore_value_data2=0,
        )
        and is_ok
    )

    print()
    gate.exception.warning("Difference for EDEP XY")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "lateral_xy_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
            paths.output / sim.get_actor("doseInXY").get_output_path("edep"),
            stats,
            tolerance=10,
            ignore_value_data2=0,
            axis="y",
        )
        and is_ok
    )

    print()
    gate.exception.warning("Difference for EDEP YZ")
    is_ok = (
        utility.assert_images(
            paths.gate_output / "lateral_yz_Protons_40MeV_sourceShapeGaussian-Edep.mhd",
            paths.output / sim.get_actor("doseInYZ").get_output_path("edep"),
            stats,
            tolerance=30,
            ignore_value_data2=0,
            axis="y",
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import opengate as gate
from opengate.tests import utility
from opengate.tests.src.test081_tle_helpers import (
    add_simple_waterbox,
    add_source,
)
import sys
import numpy as np
import itk

"""
To correctly check if the TLE applied to a high number of track per event, a brem splitting of 6 MeV electrons
is applied at a target before the waterbox entrance, with a very low
"""


def test(f1, f2, f1_err, f2_err):
    img1 = itk.imread(f1)
    img2 = itk.imread(f2)
    img1_err = itk.imread(f1_err)
    img2_err = itk.imread(f2_err)
    img1_arr = np.sum(itk.GetArrayFromImage(img1))
    img2_arr = np.sum(itk.GetArrayFromImage(img2))
    img1_err_arr = np.sum(itk.GetArrayFromImage(img1_err))
    img2_err_arr = np.sum(itk.GetArrayFromImage(img2_err))

    tot_err = np.sqrt((img1_arr * img1_err_arr) ** 2 + (img2_arr * img2_err_arr) ** 2)
    print("dose actor:", img1_arr, "+-", img1_err_arr * img1_arr, "Gy")
    print("TLE dose actor:", img2_arr, "+-", img2_err_arr * img2_arr, "Gy")
    if np.abs(img2_arr - img1_arr) < 4 * tot_err:
        return True


def main(argv):
    paths = utility.get_default_test_paths(__file__, output_folder="test081_tle")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.random_seed = "auto"
    sim.output_dir = paths.output
    sim.progress_bar = True
    sim.number_of_threads = 1
    if len(argv) > 1:
        sim.number_of_threads = int(argv[1])

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    gcm3 = gate.g4_units.g / gate.g4_units.cm3

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # create voxelized waterbox
    waterbox = add_simple_waterbox(sim)

    # save voxelized versions of the waterbox
    # voxelize_waterbox(sim, paths.data / "test081_tle")

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.global_production_cuts.all = 0.001 * mm
    sim.physics_manager.set_max_step_size("waterbox", 0.1 * mm)
    sim.physics_manager.set_user_limits_particles("gamma")

    sim.volume_manager.material_database.add_material_weights(
        "Tungsten",
        ["W"],
        [1],
        19.3 * gcm3,
    )

    # default source for tests
    source = add_source(sim, n=2500, energy=2.5 * MeV)
    source.particle = "e-"

    target = sim.add_volume("Box", "target")
    target.size = [10 * cm, 10 * cm, 2 * mm]
    target.material = "Tungsten"
    target.translation = [0, 0, -waterbox.size[2] / 2 - 1 * mm]
    target.color = [0.6, 0.8, 0.4, 0.9]

    region_linac = sim.physics_manager.add_region(name=f"{target.name}_region")
    region_linac.associate_volume(target)
    # set the brem splitting
    s = f"/process/em/setSecBiasing eBrem {region_linac.name} {100} 50 MeV"
    sim.g4_commands_after_init.append(s)

    # add tle dose actor
    tle_dose_actor = sim.add_actor("TLEDoseActor", "tle_dose_actor")
    tle_dose_actor.output_filename = "test081_tle_6_brem_split.mhd"
    tle_dose_actor.attached_to = waterbox
    tle_dose_actor.dose_uncertainty.active = True
    tle_dose_actor.dose.active = True
    tle_dose_actor.size = [1, 1, 1]
    tle_dose_actor.spacing = [x / y for x, y in zip(waterbox.size, tle_dose_actor.size)]
    tle_dose_actor.tle_threshold_type = "max range"
    tle_dose_actor.tle_threshold = 10 * mm
    print(f"TLE Dose actor pixels : {tle_dose_actor.size}")
    print(f"TLE Dose actor spacing : {tle_dose_actor.spacing} mm")
    s = f"/process/eLoss/CSDARange true"
    sim.g4_commands_before_init.append(s)

    # add conventional dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.output_filename = "test081_6_brem_split.mhd"
    dose_actor.attached_to = waterbox
    dose_actor.dose_uncertainty.active = True
    dose_actor.dose.active = True
    dose_actor.size = [1, 1, 1]
    dose_actor.spacing = [x / y for x, y in zip(waterbox.size, dose_actor.size)]
    print(f"Dose actor pixels : {dose_actor.size}")
    print(f"Dose actor spacing : {dose_actor.spacing} mm")
    print(f"Dose actor size : {waterbox.size} mm")

    print(f"TLE Dose actor size : {waterbox.size} mm")

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()
    f1 = dose_actor.dose.get_output_path()
    f2 = tle_dose_actor.dose.get_output_path()

    f1_bis = dose_actor.dose_uncertainty.get_output_path()
    f2_bis = tle_dose_actor.dose_uncertainty.get_output_path()

    # print results at the end
    print(stats)

    is_ok = test(f1, f2, f1_bis, f2_bis)
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main(sys.argv)
